#!/usr/bin/env python

__author__    = "Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


'''This examples shows how to use the rs.Filesystem API
   with the HTTP file adaptor.

   If something doesn't work as expected, try to set
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.
'''

import sys

import radical.saga as rs


def main():

    try:
        # open file on a remote web server - WARNING: size is ~ 800 MB
        remote_file = rs.filesystem.File('http://hgdownload.cse.ucsc.edu/goldenPath/hg19/bigZips/hg19.2bit')

        # copy the remote file to /tmp/ on the local machine
        remote_file.copy('file://localhost/tmp/', flags=rs.filesystem.OVERWRITE)
        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

__author__    = 'RADICAL @ Rutgers'
__copyright__ = 'Copyright 2012-2013, The SAGA Project'
__license__   = 'MIT'


'''
This examples shows how to run groups of jobs using the 'local' file adaptor.
This example uses job containers for simplified and optimized bulk job handling.

Job container can be used to model dependencies between groups of different
jobs, e.g., in workflow scenarios. In this example, we execute 'num_job_groups'
containers of jobs_per_group' number of parallel jobs sequentially::

    C1[j1,j2,j3,j4,...] -> C2[j1,j2,j3,j4,...] -> C3[j1,j2,j3,j4,...] -> ...

Depending on the adaptor implementation, using job containers can be quite
advantageous in terms of call latency. Some adaptors implement special bulk
operations for container management, which makes them generally much faster than
iterating over and operating on individual jobs.
'''

import sys
import random

import radical.saga as rs


URL = 'slurm://locahost/'


# ------------------------------------------------------------------------------
def main():

    # number of job 'groups' (containers) and of jobs per group
    num_job_groups = 10
    jobs_per_group =  2  # check slurm limits!

    current = None

    try:
        # all jobs in this example are running on the same job service
        service = rs.job.Service(URL)

        # create and populate our containers
        containers = list()
        for c in range(0, num_job_groups):

            # create containers
            containers.append(rs.job.Container())

            # add jobs to container.
            for j in range(0, jobs_per_group):
                jd = rs.job.Description()
                jd.executable  = '/bin/sleep'
                jd.arguments   = ['10']
                jd.name        = ['job.%02d.%03d' % (c, j)]
                j = service.create_job(jd)
                containers[c].add(j)

        # execute the containers sequentially
        for c in range(0, num_job_groups):
            print('run container %s ' % c)
            containers[c].run()

            current = containers[c]  # see exception handling

            for j in containers[c].get_tasks():
                print('%s: %s: %s' % (j.name, j.id, j.state))

            print(containers[c].get_states ())
            containers[c].wait()
            print(containers[c].get_states ())

            # all jobs in the container finished - print some job infos
            for job in containers[c].jobs:
                print('  * %s: %s (%s) @%s %s - %s' \
                  % (job.id, job.state, job.exit_code, job.execution_hosts,
                     job.started, job.finished))

            print()

        service.close()
        return 0

    except rs.SagaException as ex:

        print('An exception occured: %s ' % ((str(ex))))
        # get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(' *** %s' % ex.traceback)
        return -1


    finally:

        # make sure we leave no jobs behind
        if current is not None:
            print('\ncancel current container: %s' % current)
            current.cancel()
            for job in current.jobs:
                print('  * %s: %s (%s) @%s %s - %s' \
                  % (job.id, job.state, job.exit_code, job.execution_hosts,
                     job.started, job.finished))
            print()


# ------------------------------------------------------------------------------
if __name__ == '__main__':

    sys.exit(main())


# ------------------------------------------------------------------------------





__author__    = "RADICAL-SAGA Development Team"
__copyright__ = "Copyright 2013, RADICAL"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#

from .version    import *
from .constants  import *

from .task       import Task, Container
from .attributes import Attributes, Callback
from .session    import Session, DefaultSession
from .context    import Context
from .url        import Url

from .exceptions import SagaException
from .exceptions import NotImplemented
from .exceptions import IncorrectURL
from .exceptions import BadParameter
from .exceptions import AlreadyExists
from .exceptions import DoesNotExist
from .exceptions import IncorrectState
from .exceptions import PermissionDenied
from .exceptions import AuthorizationFailed
from .exceptions import AuthenticationFailed
from .exceptions import Timeout
from .exceptions import NoSuccess

from .           import job
from .           import filesystem
from .           import replica
from .           import advert
from .           import resource
# from .           import messages

from .           import utils


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .base       import CPIBase
from .decorators import CPI_SYNC_CALL  as SYNC
from .decorators import CPI_ASYNC_CALL as ASYNC


class Context (CPIBase) :
    
    @SYNC
    def init_instance         (self, type)    : pass

    @SYNC
    def _initialize           (self, session) : pass

    @SYNC
    def _get_default_contexts (self, session) : pass








__author__    = "Andre Merzky, Ashley Z, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" libcloud based EC2 resource adaptor """

import re
import os

from ...                   import exceptions as rse
from ...                   import url        as rs_url
from ...                   import session    as rs_session
from ...                   import context    as api_context
from ...                   import resource   as api_resource
from ..                    import base       as a_base
from ..cpi                 import context    as cpi_context
from ..cpi                 import resource   as cpi_resource
from ..cpi                 import decorators as cpi_decs
from ...resource           import constants  as c

c.ANY = c.COMPUTE | c.STORAGE

SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# ------------------------------------------------------------------------------
# the adaptor info
#
_ADAPTOR_NAME          = "radical.saga.adaptors.ec2_resource"
_ADAPTOR_SCHEMAS       = ["ec2", "ec2_keypair", "openstack", "eucalyptus",
                          "euca", "aws", "amazon", "http", "https"]
_ADAPTOR_OPTIONS       = []

# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "rdes_attributes"  : [c.RTYPE         ,
                          c.TEMPLATE      ,
                          c.IMAGE         ,
                          c.DYNAMIC       ,
                          c.MACHINE_OS    ,
                          c.MACHINE_ARCH  ,
                          c.SIZE          ,
                          c.MEMORY        ,
                          c.ACCESS       ],
    "res_attributes"   : [c.RTYPE         ,
                          c.TEMPLATE      ,
                          c.IMAGE         ,
                          c.MACHINE_OS    ,
                          c.MACHINE_ARCH  ,
                          c.SIZE          ,
                          c.MEMORY        ,
                          c.ACCESS       ],
    "metrics"          : [c.STATE,
                          c.STATE_DETAIL],
    "contexts"         : {"ec2"         : "EC2 ID and Secret",
                          "ec2_keypair" : "ec2 keypair for node access"}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "cfg_options"      : _ADAPTOR_OPTIONS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : """
        The EC2 resource adaptor. This adaptor interacts with a variety of
        IaaS backends via the Apache LibCloud.  It also provides EC2 related
        context types.
        """,
    "example": "examples/resource/amazon_ec2.py",
    "schemas"          : {"ec2"         : "Amacon EC2 key/secret",
                          "ec2_keypair" : "Amacon EC2 keypair name"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO          = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.1",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "cpis"             : [
                             {
                                 "type"  : "radical.saga.Context",
                                 "class" : "EC2Keypair"
                             },
                             {
                                 "type"  : "radical.saga.resource.Manager",
                                 "class" : "EC2ResourceManager"
                             },
                             {
                                 "type"  : "radical.saga.resource.Compute",
                                 "class" : "EC2ResourceCompute"
                             },
                         ]
}


###############################################################################
# The adaptor class
class Adaptor (a_base.Base):
    """

    **Known Limitations, Notes**

    1) EC2 reports the VM instance to be 'Running' when it starts booting -- at
    that point the ssh login is not yet functional, and job service instance
    creation will thus fail.  I don't see a simple way to inspect the internal
    state (apart from trial/error which is not implemented), so the application
    needs to cater for that, for example by re-trying to connect.

    2) use `export LIBCLOUD_DEBUG=/dev/stderr` for debugging

    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # for id parsing
        self.id_re = re.compile ('^\[(.*)\]-\[(.*?)\]$')

        self._default_contexts = list()


    # --------------------------------------------------------------------------
    #
    def sanity_check (self) :

        try :
            from ...utils import misc as rsutm

            # get the libclound modules.  Note that the non-empty fromlist
            # forces Python to include the actually specified module, not only
            # the top level libcloud.  Oh Python...
            self.lc   = __import__ ('libcloud',                   fromlist=[''])
            self.lcct = __import__ ('libcloud.compute.types',     fromlist=[''])
            self.lccp = __import__ ('libcloud.compute.providers', fromlist=[''])

            # we only tested against libcloud 0.12.4, so we require that version
            # for now, at least
            if  rsutm.normalize_version (self.lc.__version__) < \
                rsutm.normalize_version ('0.12.4')       :
                raise rse.NoSuccess ("Libcloud version >=0.12.4 required")


            # 0.12.4 does not support keypair footprint inspection, so we cannot
            # create ssh keys from keypair names.  We check for that feature
            # here, and issue a warning otherwise
            # for now, at least
            if  rsutm.normalize_version (self.lc.__version__) < \
                rsutm.normalize_version ('0.12.5')       :
                self._logger.warning ("Libcloud version does not allow keypair "
                                      "inspection -- cannot associate ssh keys "
                                      "with ec2 keypairs")
                self.lc_has_footprint = False
            else :
                self.lc_has_footprint = True

            # we pick EC2 environment parameters for default settings
            self._EC2_URL        = os.environ.get ('EC2_URL'       , "")
            self._EC2_SECRET_KEY = os.environ.get ('EC2_ACCESS_KEY', "")
            self._EC2_ACCESS_KEY = os.environ.get ('EC2_SECRET_KEY', "")


        except Exception as e :
            self._logger.warning ("Could not load libcloud module, "
                                  "disable EC2 resource adaptor")
            self._logger.warning (str(e))
            raise rse.NoSuccess ("Cannot load libcloud") from e


    # --------------------------------------------------------------------------
    #
    def _get_default_contexts (self) :

        if  None is self._default_contexts :

            self._default_contexts = list()

            # no default keypair in ec2 -- but lets see if we have default access
            # information
            if  not self._EC2_URL        and \
                not self._EC2_ACCESS_KEY and \
                not self._EC2_SECRET_KEY :
                # no default access info...
                return []

            # ok, lets pick up a default context from the EC2 default env vars
            ctx = api_context.Context ('ec2')

            if self._EC2_URL        : ctx.server   = self._EC2_URL
            if self._EC2_ACCESS_KEY : ctx.user_id  = self._EC2_ACCESS_KEY
            if self._EC2_SECRET_KEY : ctx.user_key = self._EC2_SECRET_KEY

            self._default_contexts.append (ctx)

        return self._default_contexts


    # --------------------------------------------------------------------------
    #
    def parse_id (self, id) :
        # split the id '[manager-url]-[resource-url]' in its parts, and return
        # them.

        match = self.id_re.match (id)

        if  not match or len (match.groups()) != 2 :
            raise rse.BadParameter ("Cannot parse resource id '%s'" % id)

        return (rs_url.Url (match.group(1)), str (match.group (2)))


    # --------------------------------------------------------------------------
    #
    def connect (self, session=None, url=None) :

        if  not session :
            session = rs_session.Session (default=True)

        ctx_id  = None
        ctx_key = None
        ctx_url = None
        error   = None

        ec2_url = None

        for ctx in session.contexts :

            if  ctx.type.lower () == 'ec2' :
                ctx_id  = ctx.user_id
                ctx_key = ctx.user_key
                ctx_url = rs_url.Url (ctx.server)

                if  not ctx_url or not str(ctx_url) :
                    ctx_url = rs_url.Url (self._EC2_URL)

                if  url :
                    ec2_url = url
                else :
                    ec2_url = ctx_url

                # ec2_url may still be empty!

                driver = None
                if  not ec2_url                     or \
                    ec2_url.scheme  == 'ec2'        or \
                    ec2_url.scheme  == 'aws'        or \
                    ec2_url.scheme  == 'amazon'        :
                    driver  = self.lccp.get_driver (self.lcct.Provider.EC2)
                    backend = 'aws'
                elif ec2_url.scheme == 'eucalyptus' or \
                    ec2_url.scheme  == 'euca'       or \
                    ec2_url.scheme  == 'http'       or \
                    ec2_url.scheme  == 'https'      or \
                    ec2_url.scheme  == 'openstack'     :
                    driver = self.lccp.get_driver(self.lcct.Provider.EUCALYPTUS)
                    backend = 'euca'
                else :
                    error = "URL schema not supported by adaptor(%s)" % ec2_url
                    next


                if  backend  == 'aws' :
                    self._logger.debug ("aws backend")
                    conn = driver (ctx_id, ctx_key)
                    # do we need URL details for different availability zones?

                elif backend == 'euca' :
                    self._logger.debug ("eucalyptus backend")
                    conn = driver (ctx_id, ctx_key,
                                 # secure = True, # how do we know?
                                   host   = ctx_url.host,
                                   port   = ctx_url.port,
                                   path   = ctx_url.path)
                else :
                    error = "only EC2 supported (not %s)" % ec2_url
                    next

                return conn, backend

        # no luck, didn't get a valid connection...
        if  error :
            raise rse.BadParameter (error)

        # no particular context failed -- raise generic exception
        raise rse.BadParameter("no valid EC2 credentials found (ec2 url='%s')"
                              % ec2_url)


###############################################################################
#
class EC2Keypair (cpi_context.Context) :

    """
    This context points to an EC2 keypair which is used to contextualize VM
    instances.

    The context can be used in two ways, depending on the specified keys:

    Version 1: reference an existing (uploaded) keypair:
      - `Token`  : name of keypair to be used  (required)
      - `UserID` : username on VM instance     (optional, default 'root')
      - `Server` : authentication server host  (optional, default for Amazon)

    A 'DoesNotExist' exception will be raised if the referenced keypair does not
    exist, and the context will not be added to the session.  An attempt to
    connect to a VM with an invalid keypair (i.e.  a keypair not registered upon
    VM creation), or with an invalid user id, will result in an
    'AuthorizationDenied' exception.


    Version 2: create (upload) a new keypair, and the use it
      - `Token`  : name of keypair to create   (required)
      - `UserKey`: private or public  ssh key  (required)
      - `UserID` : username on VM instance     (optional, default 'root')
      - `Server` : authentication server host  (optional, default for Amazon)

    When used in this version, the adaptor will attempt to create an EC2 keypair
    with the given name (`Token`), by uploading the public ssh key.  On success,
    the `UserKey` attribute will then be removed from the context, to avoid
    repeated uploads on re-use, and the context will behave as in Version 1.  If
    a keypair with the given name already exists, an 'AlreadyExists' exception
    is raised.  All other errors with result in a `NoSuccess` exception.  If any
    error occurs, the context will not be added to the session.

    The `UserKey` attribute can point to either the public or private key of the
    ssh keypair -- the RADICAL-SAGA implementation will internally complete the
    respective other key (public key file names are expected to be derived from
    the private key, by appending the suffix `.pub`).

    *Note* that `Token` is not a standard :class:`saga.Context` attribute -- at
    this point, there seems to exist no suitable attribute for that
    functionality.  The `Token` attribute may be officially included into the
    list of context attributes at some point.


    Known Limitations:

    1) For a given EC2 keypair name, we should fetch the respective key
    footprint with `conn.ex_describe_keypairs('self.api.target')`, then sift
    through all public ssh keys we can find, and see if one matches that
    footprint.  If one does, we should add that respective ssh key to the
    session, so that it can be used for host access authentication.

    Alas, the `ex_describe_keypairs()` call is faulty and does not return
    footprints, and so we have no chance really to find the respective ssh key.
    We thus need to rely on the user to add the respective ssh key to the
    session on her own.

    I have filed a bug report with libcloud[1], lets see what happens.  At this
    point, this context adaptor does basically nothing than hosting the EC2
    keypair for session authentication.

    Update: It turns out that EC2 uses a non-standard and (worse) non-specified
    way to compute the fingerprints, and we can't use those to select local
    keys.  There apparently exists a feature-request to AWS to switch to normal
    ssl MD5 digests [2], but it is unclear if and when that happens.  That
    severely impacts our ability to choose a suitable ssh key for VM access, and
    we must rely on the user to provide a suitable one, either in the
    ec2_keypair context, or in a separate ssh context.


    [1] `https://issues.apache.org/jira/browse/LIBCLOUD-326`
    [2] `https://forums.aws.amazon.com/message.jspa?messageID=386571`

    """


    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        _cpi_base = super  (EC2Keypair, self)
        _cpi_base.__init__ (api, adaptor)

        self.lcct = self._adaptor.lcct
        self.lccp = self._adaptor.lccp


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, type) :

        if  not type.lower () == 'ec2'         and \
            not type.lower () == 'ec2_keypair' :
            raise rse.BadParameter("ec2 adaptor only handles 'ec2' and "
                                   "'ec2_keypair' contexts")

        self._type = type

        return self


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def _initialize (self, session) :

        # nothing to do for simple ec2 id/secret containers
        # FIXME: we could in principle validate validity...
        if  self._type.lower () == 'ec2' :
            return

        # we first attempt to create an ssh context from the keypair
        # context -- this will take care of all eventual key checks etc.
        ssh_context = api_context.Context ('ssh')

        ssh_context.user_id   = self._api ().user_id
        ssh_context.user_key  = self._api ().user_key
        ssh_context.user_cert = self._api ().user_cert
        ssh_context.user_pass = self._api ().user_pass

        # contexts are verified on session.add_context -- to force that
        # verification we use a temporary session.  This will check if the ssh
        # key exists, etc.
        tmp_session = rs_session.Session (default=False)
        tmp_session.add_context (ssh_context)


        # we have an ec2_keypair context.  We need to find an ec2  context in
        # the session though which we can use to talk to the backend
        self.ec2_id  = None
        self.ec2_key = None

        for ctx in session.contexts :
            if  ctx.type.lower () == 'ec2' :
                self.ec2_id  = ctx.user_id
                self.ec2_key = ctx.user_key
                break  # only need one of those...

        if  not self.ec2_id or not self.ec2_key :
            raise rse.AuthenticationFailed \
                  ("no ec2 context -- cannot initialize ec2_keypair")

        if  not self._api ().token :
            raise rse.BadParameter \
                  ("`ec2_keypair` context must specify keypair name as `token`")

      # # valid context, connect to backend
      # # FIXME: use 'Server' if defined
      # conn, backend = self._adaptor.connect (session)
      #
      # # check if given keypair exists.  We only can do that by inspecting it,
      # # and capturing an eventual exception.
      # keypair = None
      # token   = self._api ().token
      # ssh_id  = ssh_context.user_id
      # key     = ssh_context.user_key
      # cert    = ssh_context.user_cert
      # keypass = ssh_context.user_pass
      #
      # # With theese information, attempt to verify or upload the keypair.
      # upload  = False
      # try:
      #     # try to find it
      #     print('describe %s' % token)
      #     keypair = conn.ex_describe_keypairs (token)
      #
      #     if  not keypair               or \
      #         not 'keyName' in keypair  or \
      #         not keypair['keyName']       :
      #
      #         self._logger.info ("keypair check nok: %s" % (keypair))
      #         upload = True
      #
      #     else :
      #         self._logger.info ("keypair check ok: %s" % (keypair))
      #
      # except Exception as e :
      #
      #     self._logger.error ("keypair check : %s" % (e))
      #
      #     # FIXME: actually, key management only seems to work for AWS/EC2
      #     # proper (ex_describe_keypairs is documented as EC2 only by
      #     # libcloud), so we only raise a warning on errors for other
      #     # backends.
      #
      #     if  backend == 'aws' :
      #
      #         if str(e).startswith ("InvalidKeyPair.NotFound") :
      #
      #             if  not key :
      #                 raise BadParameter \
      #                       ("'ec2_keypair' not found: %s" % e)
      #
      #             # keypair not found, but we have a key and can register it
      #             upload = True
      #
      #         else :
      #             raise BadParameter \
      #                   ("'ec2_keypair' invalid: %s" % e)
      #
      #
      # # upload keypair if we did not find it, and have something to upload.
      # if  upload and key :
      #
      #     self._logger.info ("import new keypair %s : %s" % (token, key))
      #
      #     try :
      #         keypair = conn.ex_import_keypair (token, key)
      #         self._logger.info ("keypair upload gave %s" % keypair)
      #
      #
      #     except Exception as e :
      #         raise BadParameter \
      #               ("'ec2_keypair' not imported: %s" % e)
      #
      #     # import worked -- we don't need to import again, so unset the
      #     # user_key attribute
      #     self._api ().user_key  = None
      #     self._api ().user_cert = None
      #
      #
      # # did not find it, and have nothing to upload?!
      # elif upload and not key :
      #     self._logger.error ("no 'UserKey' for ec2_keypair '%s'" % token)
      #
      # # keypair found all right
      # else :
      #     self._logger.info ("validated ec2_keypair '%s'" % token)


        # we add the thusly derived ssh context to the session which originally
        # contained our ec2_keypair context.  We do that even for failed
        # uploads, in the hope that some out-of-band setup kicks in.
        session.add_context (ssh_context)


###############################################################################
#
class EC2ResourceManager (cpi_resource.Manager) :
    """

    **EC2_URLs:**


    AWS Generic access point           https://ec2.amazonaws.com/
    AWS US East (Northern Virginia)    https://ec2.us-east-1.amazonaws.com/
    AWS US West (Oregon)               https://ec2.us-west-2.amazonaws.com/
    AWS US West (Northern California)  https://ec2.us-west-1.amazonaws.com/
    AWS EU (Ireland)                   https://ec2.eu-west-1.amazonaws.com/
    AWS Asia Pacific (Singapore)       https://ec2.ap-southeast-1.amazonaws.com/
    AWS Asia Pacific (Tokyo)           https://ec2.ap-northeast-1.amazonaws.com/
    AWS South America (Sao Paulo)      https://ec2.sa-east-1.amazonaws.com/

    OSDC                               euca://api.opensciencedatacloud.org:8773\
                                                        /sullivan/services/Cloud

    **Known Limitations**

    1) the EC2 backend reports a VM to be in `ACTIVE` state as soon as it begins
       to boot up (From that point in time on, the VM indeed consumes resources,
       and is billed against the user's account).  The VM is at that point,
       however, not yet reachable, as the boot process needs to be completed
       bevore the ssh daemon (or any other service) can operate correctly.
       Furthermore, the VM's IP address may not yet be registered in DNS,
       further limiting reachability of the machine.  Finally, it seems that
       also the VM contextualization is at this point not completed, so that for
       example an ssh connection may fail due to unregistered keys.

       There does not seem to exist and universal way to wait for the boot,
       registration and contextualization completion.  Retrying to connect is
       one option, but (a) the connection mechanism depends on the VM's OS image
       configuration, and (b) a connection failure due to incomplete system
       startup is indistinguishable from any 'real' system error.

       For the time being it is left to the application level on how to handle
       that problem, and this adaptor will simply report the state as returned
       by EC2.
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (EC2ResourceManager, self)
        self._cpi_base.__init__ (api, adaptor)

        self.lcct = self._adaptor.lcct
        self.lccp = self._adaptor.lccp


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, session) :

        self.url     = rs_url.Url (url)  # deep copy
        self.session = session

        # internale (cached) registry of available resources
        self.templates         = []
        self.templates_dict    = {}
        self.images            = []
        self.images_dict       = {}
        self.access            = {}
        self.access[c.COMPUTE] = []
        self.access[c.STORAGE] = []
        self.access[c.ANY]     = []

        self.conn, self.backend = self._adaptor.connect (session, url)

        self.templates = []
        self.images    = []

        # FIXME: we could pre-fetch existing resources right now...


    # --------------------------------------------------------------------------
    #
    def _refresh_templates (self, pattern=None) :

        self.templates      = []
        self.templates_dict = {}

        for template in self.conn.list_sizes (pattern) :

            self.templates_dict   [template.name] = template
            self.templates.append (template.name)


    # --------------------------------------------------------------------------
    #
    def _refresh_images (self, uid=None) :

        self.images      = []
        self.images_dict = {}

        if uid:
            pattern = [uid]
        else:
            pattern = None

        for image in self.conn.list_images (ex_image_ids=pattern) :

            if  image.id.startswith ('ami-') :

                self.images_dict   [image.id] = image
                self.images.append (image.id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def acquire (self, rd) :

        if  not self.conn :
            raise rse.IncorrectState ("not connected to backend")

        if  rd.rtype != c.COMPUTE:
            raise rse.BadParameter ("can only acquire compute resources.")


        # check if a any 'ec2_keypair' context is known.  If so, use its
        # 'keypair'
        # attribute as keypair name for node creation
        token = ''
        for context in self.session.contexts  :
            if  context.type.lower () == 'ec2_keypair' :
                token = context.token
                self._logger.info ("using '%s' as ec2 keypair" % token)

        resource_info = None

        # check that only supported attributes are provided
        for attribute in rd.list_attributes():
            if attribute not in _ADAPTOR_CAPABILITIES["rdes_attributes"]:
                msg = "'resource.Description.%s' unsupported by this adaptor" \
                                                                     % attribute
                raise rse.BadParameter._log (self._logger, msg)


        # we only support template defined instances right now
        # FIXME: should be able to select suitable template from given
        # resource attributes
        if  not rd.template :
            raise rse.BadParameter ("no 'template' in resource description")

        # we also need an OS image
        if  not rd.image :
            raise rse.BadParameter ("no 'image' in resource description")

        # and we don't support any other attribute right now
        if  rd.dynamic      or rd.start        or \
            rd.end          or rd.duration     or \
            rd.machine_os   or rd.machine_arch or \
            rd.access       or rd.memory       :
            raise rse.BadParameter ("resource descriptions only supports "
                                    "'template' and 'image' attributes "
                                    "right now")

        try :

            # make sure template and image are valid, and get handles
            if  rd.template not in self.templates_dict :
                self._refresh_templates (rd.template)

            if  rd.image not in self.images_dict :
                self._refresh_images (uid=rd.image)


            # FIXME: interpret / verify size

            # user name as id tag
            import getpass
            cid = getpass.getuser()
            _c   = self.conn

            # create/use the saga-sg security group which allows ssh access
            try:
                ret = _c.ex_create_security_group('saga-sg','SAGA', None)
                ret = _c.ex_get_security_groups(group_names=['saga-sg'])
                gid = ret[0].id
                ret = _c.ex_authorize_security_group_ingress(gid, 22, 22,
                                                         cidr_ips=['0.0.0.0/0'])
                ret = _c.ex_authorize_security_group_egress (gid, 22, 22,
                                                         cidr_ips=['0.0.0.0/0'])

            except Exception as e:
                # lets hope this was a race and the group now exists...
                pass

            # it should be safe to create the VM instance now
            node = _c.create_node(name='radical.saga.resource.Compute.%s' % cid,
                                  size=self.templates_dict[rd.template],
                                  image=self.images_dict[rd.image],
                                  ex_keyname=token,
                                  ex_security_groups=['saga-sg'])

            resource_info = {'backend'              : self.backend   ,
                             'resource'             : node           ,
                             'resource_type'        : rd.rtype       ,
                             'resource_description' : rd             ,
                             'resource_manager'     : self.get_api (),
                             'resource_manager_url' : self.url       ,
                             'resource_schema'      : self.url.schema,
                             'connection'           : self.conn      }

        except Exception as e :
            # FIXME: translate errors more sensibly
            raise rse.NoSuccess ("Failed with %s" % e) from e

        if  resource_info :
            if  rd.rtype == c.COMPUTE :
                return api_resource.Compute(_adaptor=self._adaptor,
                                            _adaptor_state=resource_info)

        raise rse.NoSuccess ("Could not acquire requested resource")


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def acquire_by_id(self, rid):

        if  not self.conn :
            raise rse.IncorrectState ("not connected to backend")


        try :

            manager_url, rid_s = self._adaptor.parse_id (str(rid))

            # FIXME: interpret / verify size
            nodes  = self.conn.list_nodes (ex_node_ids=[rid_s])

            if  len (nodes) < 1 :
                raise rse.BadParameter ("Cannot find resource '%s'" % rid_s)
            if  len (nodes) > 1 :
                raise rse.BadParameter ("Cannot identify resource '%s'" % rid_s)

            node = nodes[0]

            resource_info = {'backend'              : self.backend   ,
                             'resource'             : node           ,
                             'resource_type'        : c.COMPUTE      ,
                             'resource_description' : None           ,
                             'resource_manager'     : self.get_api (),
                             'resource_manager_url' : self.url       ,
                             'resource_schema'      : self.url.schema,
                             'connection'           : self.conn      }

        except Exception as e :
            # FIXME: translate errors more sensibly
            raise rse.NoSuccess ("Failed with %s" % e) from e

        return api_resource.Compute(_adaptor=self._adaptor,
                                    _adaptor_state=resource_info)


    # --------------------------------------------------------------------------
    @SYNC_CALL
    def get_url (self) :

        return self.url


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list (self, rtype):

        if  not self.conn :
            raise rse.IncorrectState ("not connected to backend")


        ret = []

        try :

            for node in self.conn.list_nodes () :
                ret.append ("[%s]-[%s]" % (self.url, node.id))

        except Exception as e :
            # FIXME: translate errors more sensibly
            raise rse.NoSuccess ("Failed with %s" % e) from e

        return ret


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def destroy (self, id):

        node = self.acquire (id)
        node.destroy ()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list_templates (self, rtype) :

        # we support only compute templates right now
        if  rtype and not (rtype & c.COMPUTE) :
            return []

        if not len (self.templates) :
            self._refresh_templates ()

        return self.templates


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_template (self, name) :

        # FIXME
        raise rse.BadParameter ("unknown template %s" % name)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list_images (self, rtype) :

        # we support only compute images right now
        if  rtype and not (rtype & c.COMPUTE) :
            return []

        if not len (self.images) :
            self._refresh_images ()

        return self.images


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_image (self, img_id) :

        if  img_id not in self.images_dict:
            self._refresh_images (uid=img_id)

        if  img_id not in self.images_dict:
            raise rse.BadParameter ("unknown image %s" % img_id)

        descr = dict(self.images_dict[img_id].extra)

        if  'name' not in descr :
            descr['name'] = self.images_dict[img_id].name

        for key in list(descr.keys ()) :
            if  not descr[key] :
                descr.pop (key)

        return descr


###############################################################################
#
class EC2ResourceCompute (cpi_resource.Compute) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (EC2ResourceCompute, self)
        self._cpi_base.__init__ (api, adaptor)

        self.lcct = self._adaptor.lcct
        self.lccp = self._adaptor.lccp

        self.state       = c.NEW
        self.detail      = None
        self.rid         = None
        self.rtype       = None
        self.manager     = None
        self.manager_url = None


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_info, id, session):

        # eval id if given
        if  id :
            self.manager_url, self.rid = self._adaptor.parse_id (id)
            self.manager = api_resource.Manager(self.manager_url,
                                                session=session)


            if id not in self.manager.list(c.COMPUTE):

                raise rse.BadParameter ("resource '%s' not found" % id)

            cr = self.manager.acquire(id)

            self.backend     = cr._adaptor.backend
            self.resource    = cr._adaptor.resource
            self.rtype       = cr._adaptor.rtype
            self.descr       = cr._adaptor.descr
            self.manager     = cr._adaptor.manager
            self.manager_url = cr._adaptor.manager_url
            self.conn        = cr._adaptor.conn
            self.rid         = cr._adaptor.rid
            self.id          = cr._adaptor.id
            self.access      = cr._adaptor.access


        # no id -- grab info from adaptor_info
        elif adaptor_info :

            if  'backend'              not in adaptor_info or \
                'resource'             not in adaptor_info or \
                'resource_type'        not in adaptor_info or \
                'resource_description' not in adaptor_info or \
                'resource_manager'     not in adaptor_info or \
                'resource_manager_url' not in adaptor_info or \
                'connection'           not in adaptor_info    :
                raise rse.BadParameter(
                            "Cannot acquire resource, insufficient information")

            self.backend     = adaptor_info['backend']
            self.resource    = adaptor_info['resource']
            self.rtype       = adaptor_info['resource_type']
            self.descr       = adaptor_info['resource_description']
            self.manager     = adaptor_info['resource_manager']
            self.manager_url = adaptor_info['resource_manager_url']
            self.conn        = adaptor_info['connection']

            self.rid    = self.resource.id
            self.id     = "[%s]-[%s]" % (self.manager_url, self.rid)
            self.access = None


        else :
            raise rse.BadParameter("Cannot acquire resource, no id/contact")


        # FIXME: we don't actually need new state, it should be fresh at
        # this point, right?!
        self._refresh_state ()


        return self.get_api ()


    # --------------------------------------------------------------------------
    #
    def _refresh_state (self) :

        if  self.state == c.EXPIRED :
            # no need to update, state is final
            return

        try :
            # NOTE: ex_node_ids is only supported by ec2
            nodes = self.conn.list_nodes (ex_node_ids=[self.rid])

            if  not len (nodes) :
                raise rse.IncorrectState ("resource '%s' disappeared")

            if  len (nodes) != 1 :
                self._logger.warning ("Cannot identify instance %s" % self.rid)

            self.resource = nodes[0]

            if  'status' in self.resource.extra :
                self.detail = self.resource.extra['status']

            # FIXME: move state translation to adaptor
            s = self.resource.state
            if   s == self.lcct.NodeState.RUNNING    : self.state = c.ACTIVE
            elif s == self.lcct.NodeState.REBOOTING  : self.state = c.PENDING
            elif s == self.lcct.NodeState.TERMINATED : self.state = c.EXPIRED
            elif s == self.lcct.NodeState.PENDING    : self.state = c.PENDING
            elif s == self.lcct.NodeState.UNKNOWN    : self.state = c.UNKNOWN
            else                                     : self.state = c.UNKNOWN

            if  self.state == c.UNKNOWN and self.detail == 'shutting-down' :
                self.state =  c.EXPIRED

        except Exception as e :
            self._logger.error ("Could not obtain resource state (%s): %s"
                             % (self.id, e))
            self.state  =  c.UNKNOWN

        if  self.state  == c.EXPIRED :
            self.access is None
        else :
            if  len (self.resource.public_ips) :
                self.access = "ssh://%s/" % self.resource.public_ips[0]


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id (self) :

        return self.id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_rtype (self) :

        return self.rtype


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state (self) :

        return self.state


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state_detail (self) :

        return self.detail


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_access (self) :

        if  not self.access :
            self._refresh_state ()

        return self.access


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_manager      (self) :

        return self.manager


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description  (self) :

        return self.descr


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def reconfig (self):
        raise rse.NotImplemented ("This backend cannot reconfigre resources")


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def destroy (self):

        self.conn.destroy_node (self.resource)
        self.state  = c.EXPIRED
        self.detail = 'destroyed by user'


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def wait (self, state, timeout) :

        import time
        start = time.time ()

        while not self.state :
            self._refresh_state ()

        if  self.state == c.EXPIRED and \
            not  state  & c.EXPIRED :
                raise rse.IncorrectState ("resource is in final state (%s): %s"
                                        % (self.detail, self.id))

        while not (self.state & state):

            self._logger.info("wait   for resource state %s: %s"
                             % (state, self.state))

            if timeout > 0 :
                now = time.time ()

                if  (now - start > timeout) :
                    break

            elif timeout == 0 :
                break

            self._refresh_state ()

        self._logger.info("waited for resource state %s: %s"
                         % (state, self.state))
        return


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the CPI base class and a number of call decorators. """

import weakref

import radical.utils as ru

from ...exceptions import *


# ------------------------------------------------------------------------------
#
# CPI base class
#
class CPIBase(object):

    # --------------------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._session   = None
        self._adaptor   = adaptor
        self._cpi_cname = self.__class__.__name__
        self._logger    = ru.Logger('radical.saga.cpi')

        # The API object must obviously keep an adaptor instance.  If we also
        # keep an API instance ref in this adaptor base, we create a ref cycle
        # which will annoy (i.e. disable) garbage collection.  We thus use weak
        # references to break that cycle.  The inheriting classes MUST use
        # get_api() to obtain the API reference.
      # if  api :
      #     self._api   = weakref.ref (api)
      # else :
      #     self._api   = None
        self._api   = weakref.ref (api)

        # by default, we assume that no bulk optimizations are supported by the
        # adaptor class.  Any adaptor class supporting bulks ops must overwrite
        # the ``_container`` attribute (via
        # ``self._set_container(container=None)``, and have it point to the
        # class which implements the respective ``container_*`` methods.
        self._container = None


    def _set_container (self, container=None) :
        self._container = container


    def get_cpi_cname (self) :
        return self._cpi_cname


    def get_api (self) :

        # get api from weakref.  We can be quite confident that the api
        # object has *not* been garbage collected, yet, as it obviously is
        # still binding this adaptor instance.
        return self._api ()

    #   if self._api :
    #       # get api from weakref.  We can be quite confident that the api
    #       # object has *not* been garbage collected, yet, as it obviously is
    #       # still binding this adaptor instance.
    #       return self._api ()
    #   else :
    #       # no need to de-weakref 'None'
    #       return self._api


    def get_adaptor_name (self) :
        return self._adaptor.get_name ()


    def _set_session (self, session) :
        self._session = session


    def get_session (self) :
        return self._session

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, session):
        self._session = session




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import namespace  as cpi_ns
from .. import attributes as cpi_att

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# keep order of inheritance!  super() below uses MRO
class Directory (cpi_ns.directory.Directory,
                 cpi_att.Attributes) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsdirec = super  (Directory, self)
        self._cpi_nsdirec.__init__ (api, adaptor)

    @SYNC
    def init_instance           (self, url, flags, session)          : pass
    @ASYNC
    def init_instance_async     (self, url, flags, session)          : pass


    # ----------------------------------------------------------------
    #
    # advert methods
    #
    @SYNC
    def set_ttl_self            (self, ttl, ttype=None)              : pass
    @ASYNC
    def set_ttl_self_async      (self, ttl, ttype=None)              : pass

    @SYNC
    def get_ttl_self            (self, ttype)                        : pass
    @ASYNC
    def get_ttl_self_async      (self, ttype)                        : pass

    @SYNC
    def set_ttl                 (self, tgt, ttl, ttype)              : pass
    @ASYNC
    def set_ttl_async           (self, tgt, ttl, ttype)              : pass

    @SYNC
    def get_ttl                 (self, tgt, ttype)                   : pass
    @ASYNC
    def get_ttl_async           (self, tgt, ttype)                   : pass

    @SYNC
    def find_adverts            (self, name_pattern, attr_pattern,
                                 obj_type, flags, ttype)             : pass
    @ASYNC
    def find_adverts_async      (self, name_pattern, attr_pattern,
                                 obj_type, flags, ttype)             : pass









__author__    = "Andre Merzky"
__copyright__ = "Copyright 2020, The SAGA Project"
__license__   = "MIT"


""" mock file adaptor implementation """


from ...                     import exceptions as rse
from ...                     import url        as rsurl
from ...utils                import pty_mock  as rsups
from ...utils                import misc       as rsumisc
from ...                     import filesystem as api_fs
from ...adaptors             import base       as a_base
from ...adaptors.cpi         import filesystem as cpi_fs
from ...adaptors.cpi         import decorators as cpi_decs
from ...filesystem           import constants  as c


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.mock_file"
_ADAPTOR_SCHEMAS       = ["mock"]
_ADAPTOR_OPTIONS       = []

# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "metrics"          : [],
    "contexts"         : {}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "cfg_options"      : _ADAPTOR_OPTIONS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : """
        The mock file adaptor which does nothing.
        """,
    "details"          : """
        The mock file adaptor which does nothing.
        """,
    "schemas"          : {"mock"    :"do nothing"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO          = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.1",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "cpis"             : [
        {
            "type"     : "radical.saga.namespace.Directory",
            "class"    : "MockDirectory"
        },
        {
            "type"     : "radical.saga.namespace.Entry",
            "class"    : "MockFile"
        },
        {
            "type"     : "radical.saga.filesystem.Directory",
            "class"    : "MockDirectory"
        },
        {
            "type"     : "radical.saga.filesystem.File",
            "class"    : "MockFile"
        }
    ]
}


# ------------------------------------------------------------------------------
#
# use the native python facilities to create local directories
#
def _mkdir_p(path):

    try:
        os.makedirs(path)

    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass

    except Exception as e:
        raise rse.NoSuccess("mkdir failed for '%s': %s" % (path, e)) from e


###############################################################################
# The adaptor class

class Adaptor(a_base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):

        # FIXME: also check for gsissh

        pass


    # --------------------------------------------------------------------------
    #
    def get_lease_target(self, tgt):

        """
        return a URL with empty path which can be used to identify leased copy
        mocks (we don't care about the path while leasing copy mocks)
        """

        lease_tgt = rsurl.Url(tgt)
        lease_tgt.path = '/mock_file_adaptor_command_mock/'

        return lease_tgt


###############################################################################
#
class MockDirectory(cpi_fs.Directory):
    """ Implements cpi_fs.Directory """

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(MockDirectory, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.finalize(kill=True)


    # --------------------------------------------------------------------------
    #
    def _is_valid(self):

        if  not self.valid:
            raise rse.IncorrectState("this instance was closed or removed")

    # --------------------------------------------------------------------------
    #
    def _create_parent(self, cwdurl, tgt):

        dirname = rsumisc.url_get_dirname(tgt)
        ret     = None
        out     = None

        if  rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, _ = self._command(" mkdir -p '%s'\n" % (dirname),
                                        make_location=True)
            if  ret:
                raise rse.NoSuccess("failed at mkdir '%s': (%s) (%s)"
                                   % (dirname, ret, out))

        elif rsumisc.url_is_local(tgt):

            if  tgt.scheme and not tgt.scheme.lower() in _ADAPTOR_SCHEMAS:
                raise rse.BadParameter("unsupported mkdir schema (%s)" % tgt)
            _mkdir_p(dirname)

        else:
            lease_tgt = self._adaptor.get_lease_target(tgt)
            with self.lm.lease(lease_tgt, self.mock_creator, tgt) as tmp_mock:
                tmp_mock.run_sync('mkdir -p %s' % dirname)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):
        """ Directory instance constructor """

        if flags is None:
            flags = 0

        self.url     = rsurl.Url(url)  # deep copy
        self.flags   = flags
        self.session = session
        self.valid   = False           # will be set by initialize
        self.lm      = session._lease_manager

        # Use `_set_session` method of the base class to set the session object.
        # `_set_session` and `get_session` methods are provided by `CPIBase`.
        self._set_session(session)

        def _mock_creator(url):
            return rsups.PTYMock(url, self.get_session(), self._logger)
        self.mock_creator = _mock_creator

        # The dir command mock is leased, as the dir seems to be used
        # extensively in some cases.  Note that before each command, we need to
        # perform a 'cd' to the target location, to make sure we operate in the
        # right location (see self._command())

        self.initialize()

        return self.get_api()

    # --------------------------------------------------------------------------
    #
    def _command(self, command, location=None, make_location=False):

        if  not location:
            location = self.url
        else:
            location = rsurl.Url(location)

        lease_tgt = self._adaptor.get_lease_target(location)
        with self.lm.lease(lease_tgt, self.mock_creator, location) \
             as cmd_mock:

            if  make_location and location.path:
                pre_cmd = "mkdir -p %s &&" % location.path
            else:
                pre_cmd = ""

            return cmd_mock.run_sync("%s cd %s && %s"
                                     % (pre_cmd, location.path, command))


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # mock got started, found its prompt.  Now, change
        # to the initial (or later current) working directory.

        cmd  = ""
        mkl  = False    # make location on command execution
        path = self.url.path
        if not path:
            path = '.'

        if  self.flags & c.CREATE_PARENTS:
            cmd = " mkdir -p '%s' ;  cd '%s'" % (path, path)
            mkl = True
        elif self.flags & c.CREATE:
            cmd = " mkdir    '%s' ;  cd '%s'" % (path, path)
            mkl = False
        else:
            cmd = " test -d  '%s' && cd '%s'" % (path, path)
            mkl = False

        # we can't yet 'cd' to self.url if we are to create it, so we start at
        # root on the target FS.
        if mkl:
            root = rsurl.Url(self.url)
            root.path = '/'
            ret, out, _ = self._command(cmd, location=root, make_location=mkl)
        else:
            ret, out, _ = self._command(cmd, make_location=mkl)

        if  ret:
            raise rse.BadParameter("invalid dir '%s': %s" % (path, out))

        self._logger.debug("initialized directory (%s)(%s)" % (ret, out))

        self.valid = True


    # --------------------------------------------------------------------------
    #
    def finalize(self, kill=False):

        self.valid = False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open(self, url, flags):

        self._is_valid()

        adaptor_state = {"from_open": True,
                         "cwd"      : rsurl.Url(self.url)}  # deep copy

        if  rsumisc.url_is_relative(url):
            url = rsumisc.url_make_absolute(self.get_url(), url)

        return api_fs.File(url=url, flags=flags, session=self.session,
                           _adaptor=self._adaptor, _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def open_dir(self, url, flags):

        self._is_valid()

        adaptor_state = {"from_open": True,
                         "cwd"      : rsurl.Url(self.url)}  # deep copy

        if  rsumisc.url_is_relative(url):
            url = rsumisc.url_make_absolute(self.get_url(), url)

        return api_fs.Directory(url=url, flags=flags, session=self.session,
                                _adaptor=self._adaptor,
                                _adaptor_state=adaptor_state)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def change_dir(self, tgt, flags):

        cwdurl = rsurl.Url(self.url)
        tgturl = rsurl.Url(tgt)
        path   = tgturl.path
        if not path:
            path = '.'

        if  not rsumisc.url_is_compatible(cwdurl, tgturl):
            raise rse.BadParameter("target dir outside of namespace '%s': %s"
                                  % (cwdurl, tgturl))

        cmd = None

        if path in ['.', cwdurl.path]:
            self._logger.debug("change directory optimized away (%s) == (%s)"
                               % (cwdurl, tgturl))

        if  flags & c.CREATE_PARENTS:
            cmd = " mkdir -p '%s' ;  cd '%s'" % (path, path)
        elif flags & c.CREATE:
            cmd = " mkdir    '%s' ;  cd '%s'" % (path, path)
        else:
            cmd = " test -d  '%s' && cd '%s'" % (path, path)

        ret, out, _ = self._command(cmd)

        if  ret:
            raise rse.BadParameter("invalid dir '%s': %s" % (cwdurl, tgturl))

        self._logger.debug("changed directory (%s)(%s)" % (ret, out))

        self.valid = True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if  timeout:
            raise rse.BadParameter("timeout for close not supported")

        self.finalize(kill=True)


    # --------------------------------------------------------------------------
    @SYNC_CALL
    def get_url(self):

        self._is_valid()

        return rsurl.Url(self.url)  # deep copy


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self, npat, flags):

        self._is_valid()

        # FIXME: eval flags

        if  npat is None:
            npat = "."
        else:
            npat = '-d %s' % npat

        ret, out, _ = self._command(" /bin/ls -C1 %s\n" % npat)

        if  ret:
            raise rse.NoSuccess("failed to list(): (%s)(%s)"
                           % (ret, out))

        lines = [_f for _f in out.split("\n") if _f]
        self._logger.debug(lines)

        self.entries = []
        for line in lines:
            # FIXME: convert to absolute URLs?
            self.entries.append(rsurl.Url(line.strip()))

        return self.entries


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt, flags):

        self._is_valid()

        # FIXME: eval flags

        return self.copy(self.url, tgt, flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy(self, src_in, tgt_in, flags, _from_task=None):

        self._is_valid()

        # FIXME: eval flags

        cwdurl = rsurl.Url(self.url)  # deep copy
        src    = rsurl.Url(src_in)    # deep copy
        tgt    = rsurl.Url(tgt_in)    # deep copy

        if rsumisc.url_is_relative(src):
            src = rsumisc.url_make_absolute(cwdurl, src)
        if rsumisc.url_is_relative(tgt):
            tgt = rsumisc.url_make_absolute(cwdurl, tgt)

        rec_flag = ""
        if  flags & c.RECURSIVE:
            rec_flag  += "-r "

        files_copied = list()

        # if cwd, src and tgt point to the same host, we just run a mock cp
        # command on that host
        if  rsumisc.url_is_compatible(cwdurl, src) and \
            rsumisc.url_is_compatible(cwdurl, tgt):

            if  flags & c.CREATE_PARENTS:
                self._create_parent(cwdurl, tgt)

            ret, out, err = self._command(" cp %s '%s' '%s'\n"
                                         % (rec_flag, src.path, tgt.path))
            if  ret:
                raise rse.NoSuccess("copy (%s -> %s) failed (%s): %s (%s)"
                               % (src, tgt, ret, out, err))


        # src and tgt are on different hosts, we need to find out which of them
        # is local (stage_from vs. stage_to).
        else:
            # print "! mock cp"

            # for a remote target, we need to manually create the target dir (if
            # needed)
            if  flags & c.CREATE_PARENTS:
                self._create_parent(cwdurl, tgt)


            # if cwd is remote, we use stage from/to
            if  not rsumisc.url_is_local(cwdurl):

                # print "cwd remote"
                if  rsumisc.url_is_local(src) and \
                    rsumisc.url_is_compatible(cwdurl, tgt):

                    lease_tgt = self._adaptor.get_lease_target(self.url)
                    with self.lm.lease(lease_tgt, self.mock_creator, self.url)\
                        as copy_mock:
                        files_copied = copy_mock.stage_to_remote(src.path,
                                                       tgt.path, rec_flag)

                elif rsumisc.url_is_local(tgt) and \
                     rsumisc.url_is_compatible(cwdurl, src):

                    lease_tgt = self._adaptor.get_lease_target(self.url)
                    with self.lm.lease(lease_tgt, self.mock_creator, self.url)\
                        as copy_mock:
                        files_copied = copy_mock.stage_from_remote(src.path,
                                                           tgt.path, rec_flag)

                else:
                    # we cannot support the combination of URLs
                    raise rse.BadParameter("unsupported copy %s to %s" % (src, tgt))


            # if cwd is local, and src or tgt are remote, we need to actually
            # create a new pipe to the target host.  note that we may not have
            # a useful session for that!
            else:
                # rsumisc.url_is_local(cwdurl):

                if  rsumisc.url_is_local(src):

                    # need a compatible target scheme
                    if  tgt.scheme and \
                        tgt.scheme.lower() not in _ADAPTOR_SCHEMAS:
                        raise rse.BadParameter("unsupported schema (%s)" % tgt)

                    # print "from local to remote"
                    lease_tgt = self._adaptor.get_lease_target(tgt)
                    with self.lm.lease(lease_tgt, self.mock_creator, tgt) \
                        as copy_mock:
                        files_copied = copy_mock.stage_to_remote(src.path,
                                                         tgt.path, rec_flag)

                elif rsumisc.url_is_local(tgt):

                    # need a compatible source scheme
                    if  src.scheme and \
                        src.scheme.lower() not in _ADAPTOR_SCHEMAS:
                        raise rse.BadParameter("unsupported source (%s)" % src)

                    # print "from remote to local"
                    lease_tgt = self._adaptor.get_lease_target(tgt)
                    with self.lm.lease(lease_tgt, self.mock_creator, tgt) \
                        as copy_mock:
                        files_copied = copy_mock.stage_from_remote(src.path,
                                                           tgt.path, rec_flag)

                else:

                    # print "from remote to other remote -- fail"
                    # we cannot support the combination of URLs
                    raise rse.BadParameter("unsupported copy %s to %s"
                                          % (src, tgt))

        if  _from_task:
            _from_task._set_metric('files_copied', files_copied)

        return files_copied


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def link_self(self, tgt, flags):

        self._is_valid()

        # FIXME: eval flags

        return self.link(self.url, tgt, flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def link(self, src_in, tgt_in, flags, _from_task=None):

        # link will *only* work if src and tgt are on the same resource (and
        # even then may fail)
        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy
        src    = rsurl.Url(src_in)    # deep copy
        tgt    = rsurl.Url(tgt_in)    # deep copy

        if  flags & c.RECURSIVE:
            raise rse.BadParameter("'RECURSIVE' flag not  supported for link()")

        if  flags & c.CREATE_PARENTS:
            self._create_parent(cwdurl, tgt)

        # if src and tgt point to the same host, we just run a mock link
        # on that host
        if  rsumisc.url_is_compatible(cwdurl, src) and \
            rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, err = self._command(" ln -s '%s' '%s'\n"
                                         % (src.path, tgt.path))
            if ret:
                raise rse.NoSuccess("link (%s -> %s) failed (%s): %s [%s]"
                                    % (src, tgt, ret, out, err))

        # src and tgt are on different hosts, this is not supported
        else:
            raise rse.BadParameter("link unsupported on FS other than cwd")


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move_self(self, tgt, flags):

        self._is_valid()

        # FIXME: eval flags

        self.move(self.url, tgt, flags)

        # need to re-initialize for new location
        self.url   = tgt
        self.flags = flags
        self.initialize()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move(self, src_in, tgt_in, flags):

        # we handle move non-atomically, i.e. as copy/remove
        self.copy  (src_in, tgt_in, flags)
        self.remove(src_in, flags)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):

        self._is_valid()

        # FIXME: eval flags

        self.remove(self.url, flags)
        self.invalid = True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove(self, tgt_in, flags):

        self._is_valid()

        # FIXME: eval flags
        # FIXME: check if tgt remove happens to affect cwd... :-P

        cwdurl = rsurl.Url(self.url)  # deep copy
        tgt    = rsurl.Url(tgt_in)    # deep copy

        rec_flag = ""
        if  flags & c.RECURSIVE:
            rec_flag += "-r "

        if  rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, err = self._command(" rm -f %s '%s'\n"
                                         % (rec_flag, tgt.path))
            if ret:
                raise rse.NoSuccess("remove (%s) failed (%s): %s [%s]"
                               % (tgt, ret, out, err))

        # we cannot support the URL
        else:
            raise rse.BadParameter("remove of %s is not supported" % tgt)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def make_dir(self, tgt_in, flags):

        self._is_valid()

        tgt  = rsurl.Url(tgt_in)  # deep copy
        chk  = ''
        opt  = ''
        path = tgt.path

        if not path:
            path = '.'

        if flags & c.EXCLUSIVE:

            # FIXME: this creates a race condition between testing for exclusive
            # mkdir and creating the dir.
            chk = ' test -d %s && echo "RS_EXISTS" || ' % path

        if flags & c.CREATE_PARENTS:
            opt = "-p"

        ret, out, err = self._command(" %s mkdir %s '%s'"
                                     % (chk, opt, path), make_location=True)

        if ret:
            raise rse.AlreadyExists("target (%s) exists (%s)" % (tgt_in, out))

        if '_EXISTS' in out:
            raise rse.AlreadyExists("make_dir target (%s) exists" % tgt_in)

        if ret:
            raise rse.NoSuccess("make_dir (%s) error: %s [%s]"
                               % (tgt_in, out, err))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size_self(self):

        self._is_valid()

        return self.get_size(self.url)

    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size(self, tgt_in):

        self._is_valid()

        tgt = rsurl.Url(tgt_in)  # deep copy

        ret, out, err = self._command(" du -ks '%s' | xargs | cut -f 1 -d ' '\n"
                                     % tgt.path)
        if ret:
            raise rse.NoSuccess("get size for (%s) failed (%s): %s [%s]"
                                % (tgt, ret, out, err))

        try:
            return int(out) * 1024  # see '-k' option to 'du'
        except Exception as e:
            raise rse.NoSuccess("could not get file size: %s (%s)" % (out, e)) \
                  from e


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def exists_self(self):

        self._is_valid()

        return self.exists(self.url)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def exists(self, tgt_in):

        self._is_valid()

        tgt = rsurl.Url(tgt_in)  # deep copy

        ret, out, _ = self._command(" test -e '%s'" % tgt.path)

        if ret: return False
        else  : return True


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):

        self._is_valid()

        return self.is_dir(self.url)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir(self, tgt_in):

        self._is_valid()

        tgt  = rsurl.Url(tgt_in)  # deep copy
        path = tgt.path
        if not path:
            path = '.'

        ret, out, _ = self._command(" test -d '%s' && test ! -h '%s'"
                    % (path, path))

        if ret: return False
        else  : return True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_entry_self(self):

        self._is_valid()

        return self.is_entry(self.url)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_entry(self, tgt_in):

        self._is_valid()

        tgt = rsurl.Url(tgt_in)  # deep copy

        ret, out, _ = self._command(" test -f '%s' && test ! -h '%s'"
                                   % (tgt.path, tgt.path))
        if ret: return False
        else  : return True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):

        self._is_valid()

        return self.is_link(self.url)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link(self, tgt_in):

        self._is_valid()

        tgt = rsurl.Url(tgt_in)  # deep copy

        ret, out, _ = self._command(" test -h '%s'" % tgt.path)

        if ret: return False
        else  : return True


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):

        return self.is_entry_self()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file(self, tgt_in):

        return self.is_entry(tgt_in)


###############################################################################
#
class MockFile(cpi_fs.File):


    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(MockFile, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        self.finalize(kill=True)


    # --------------------------------------------------------------------------
    #
    def _is_valid(self):

        if not self.valid:
            raise rse.IncorrectState("this instance was closed or removed")


    # --------------------------------------------------------------------------
    #
    def _create_parent(self, cwdurl, tgt):

        dirname = rsumisc.url_get_dirname(tgt)

        if rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, _ = self._run_sync(" mkdir -p '%s'\n" % (dirname))
            if ret:
                raise rse.NoSuccess("mkdir failed '%s': (%s) (%s)"
                                   % (dirname, ret, out))

        elif rsumisc.url_is_local(tgt):

            if  tgt.scheme and \
                tgt.scheme.lower() not in _ADAPTOR_SCHEMAS:
                raise rse.BadParameter("unsupported mkdir schema (%s)" % (tgt))

            _mkdir_p(dirname)

        else:

            lease_tgt = self._adaptor.get_lease_target(tgt)
            with self.lm.lease(lease_tgt, self.mock_creator, tgt) as tmp_mock:
                tmp_mock.run_sync('mkdir -p %s' % dirname)


    # --------------------------------------------------------------------------
    #
    def _run_sync(self, cmd):
        """
        lease the mock, set $PWD, run command, and release mock again
        """

        lease_tgt = self._adaptor.get_lease_target(self.cwdurl)
        cwd_path  = self.cwdurl.path

        with self.lm.lease(lease_tgt, self.mock_creator, self.cwdurl) as mock:

            return mock.run_sync("cd %s && %s\n" % (cwd_path, cmd))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):

        # FIXME: eval flags!
        if  flags is None:
            flags = 0

        self._logger.info("init_instance %s" % url)

        if  'from_open' in adaptor_state and adaptor_state['from_open']:

            # comes from job.service.create_job()
            self.url         = rsurl.Url(url)  # deep copy
            self.flags       = flags
            self.session     = session
            self.valid       = False  # will be set by initialize
            self.lm          = session._lease_manager
            self.cwdurl      = rsurl.Url(adaptor_state["cwd"])
            self.cwd         = self.cwdurl.path

            if  rsumisc.url_is_relative(self.url):
                self.url = rsumisc.url_make_absolute(self.cwd, self.url)

        else:

            if  rsumisc.url_is_relative(url):
                raise rse.BadParameter("cannot handle relative URL (%s)" % url)

            self.url         = url
            self.flags       = flags
            self.session     = session
            self.valid       = False  # will be set by initialize
            self.lm          = session._lease_manager

            self.cwd         = rsumisc.url_get_dirname(url)
            self.cwdurl      = rsurl.Url(url)  # deep copy
            self.cwdurl.path = self.cwd

            if  not self.flags:
                self.flags = 0

        # Use `_set_session` method of the base class to set the session object
        # `_set_session` and `get_session` methods are provided by `CPIBase`.
        self._set_session(session)

        def _mock_creator(url):
            return rsups.PTYmock(url, self.get_session(), self._logger)
        self.mock_creator = _mock_creator

     #  # self.mock is also a leased mock -- for File, it does not have any
     #  # state, really.
     #  # FIXME: get ssh Master connection from _adaptor dict
     #  lease_tgt  = self._adaptor.get_lease_target(self.url)
     #  self.mock = self.lm.lease(lease_tgt, self.mock_creator, self.url)
     #  # TODO : release mock
     #
     # self.mock.set_initialize_hook(self.initialize)
     # self.mock.set_finalize_hook  (self.finalize)

        self.initialize()

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # mock got started, found its prompt.  Now, change
        # to the initial (or later current) working directory.

        cmd = ""
        dirname = rsumisc.url_get_dirname(self.url)

        if  self.flags & c.CREATE_PARENTS:
            cmd = " mkdir -p '%s'; touch '%s'"   % (dirname, self.url.path)
            self._logger.info("mkdir '%s'; '%s'" % (dirname, self.url.path))

        elif self.flags & c.CREATE:
            cmd = " touch '%s'" % (self.url.path)
            self._logger.info("touch %s" % self.url.path)

        else:
            cmd = " true"


        if  self.flags & c.READ:
            cmd += "; test -r '%s'" % (self.url.path)

        if  self.flags & c.WRITE:
            cmd += "; test -w '%s'" % (self.url.path)

        ret, out, _ = self._run_sync(cmd)

        if  ret:
            if  self.flags & c.CREATE_PARENTS:
                raise rse.BadParameter("cannot open/create: '%s' - %s"
                                  % (self.url.path, out))
            elif self.flags & c.CREATE:
                raise rse.BadParameter("cannot open/create: '%s' - %s"
                                  % (self.url.path, out))
            else:
                raise rse.DoesNotExist("File does not exist: '%s' - %s"
                                  % (self.url.path, out))

        self._logger.info("file initialized (%s)(%s)" % (ret, out))

        self.valid = True


    # --------------------------------------------------------------------------
    #
    def finalize(self, kill=False):

        self.valid = False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if timeout:
            raise rse.BadParameter("timeout for close not supported")

        self.finalize(kill=True)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):

        self._is_valid()

        return rsurl.Url(self.url)  # deep copy


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt_in, flags):

        self._is_valid()

        # FIXME: eval flags

        # print "copy self %s -> %s" % (self.url, tgt_in)

        cwdurl = rsurl.Url(self.cwdurl)  # deep copy
        src    = rsurl.Url(self.url)     # deep copy
        tgt    = rsurl.Url(tgt_in)       # deep copy

        if rsumisc.url_is_relative(src):
            src = rsumisc.url_make_absolute(cwdurl, src)
        if rsumisc.url_is_relative(tgt):
            tgt = rsumisc.url_make_absolute(cwdurl, tgt)

        rec_flag = ""
        if  flags & c.RECURSIVE:
            rec_flag += "-r "

        if  flags & c.CREATE_PARENTS:
            self._create_parent(cwdurl, tgt)

        # if cwd, src and tgt point to the same host, we just run a mock cp
        # command on that host
        if  rsumisc.url_is_compatible(cwdurl, src) and \
            rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, _ = self._run_sync(" cp %s '%s' '%s'\n"
                                        % (rec_flag, src.path, tgt.path))
            if  ret:
                raise rse.NoSuccess("copy (%s -> %s) failed (%s): (%s)"
                                    % (src, tgt, ret, out))


        # src and tgt are on different hosts, we need to find out which of them
        # is local (stage_from vs. stage_to).
        else:
            # print "! mock cp"

            # if cwd is remote, we use stage from/to on the existing pipe
            if  not rsumisc.url_is_local(cwdurl):

                if  rsumisc.url_is_local(src) and \
                    rsumisc.url_is_compatible(cwdurl, tgt):

                    lease_tgt = self._adaptor.get_lease_target(self.url)
                    with self.lm.lease(lease_tgt, self.mock_creator, self.url)\
                        as copy_mock:
                        files_copied = copy_mock.stage_to_remote(src.path,
                                                         tgt.path, rec_flag)

                elif rsumisc.url_is_local(tgt) and \
                     rsumisc.url_is_compatible(cwdurl, src):

                    # print "from remote to local"
                    lease_tgt = self._adaptor.get_lease_target(self.url)
                    with self.lm.lease(lease_tgt, self.mock_creator, self.url)\
                        as copy_mock:
                        files_copied = copy_mock.stage_from_remote(src.path,
                                                           tgt.path, rec_flag)

                else:
                    # print "from remote to other remote -- fail"
                    # we cannot support the combination of URLs
                    raise rse.BadParameter("unsupported copy %s to %s"
                            % (src, tgt))


            # if cwd is local, and src or tgt are remote, we need to actually
            # create a new pipe to the target host.  note that we may not have
            # a useful session for that!
            else:
                # rsumisc.url_is_local(cwdurl):

                if rsumisc.url_is_local(src):

                    # need a compatible target scheme
                    if  tgt.scheme and \
                        tgt.scheme.lower() not in _ADAPTOR_SCHEMAS:
                        raise rse.BadParameter("unsupported schema (%s)" % tgt)

                    # print "from local to remote"
                    lease_tgt = self._adaptor.get_lease_target(tgt)
                    with self.lm.lease(lease_tgt, self.mock_creator, tgt) \
                        as copy_mock:
                        files_copied = copy_mock.stage_to_remote(src.path,
                                                         tgt.path, rec_flag)

                elif rsumisc.url_is_local(tgt):

                    # need a compatible source scheme
                    if  src.scheme and \
                        src.scheme.lower() not in _ADAPTOR_SCHEMAS:
                        raise rse.BadParameter("unsupported schema (%s)" % src)

                    # print "from remote to local"
                    lease_tgt = self._adaptor.get_lease_target(tgt)
                    with self.lm.lease(lease_tgt, self.mock_creator, tgt) \
                        as copy_mock:
                        copy_mock.stage_from_remote(src.path,
                                                     tgt.path, rec_flag)
                else:

                    # we cannot support two remote URLs
                    raise rse.BadParameter("unsupported copy %s to %s"
                                          % (src, tgt))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def link_self(self, tgt_in, flags, _from_task=None):

        # link will *only* work if src and tgt are on the same resource (and
        # even then may fail)
        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy
        src    = rsurl.Url(self.url)  # deep copy
        tgt    = rsurl.Url(tgt_in)    # deep copy

        if  flags & c.RECURSIVE:
            raise rse.BadParameter("'RECURSIVE' flag not  supported for link()")

        if  flags & c.CREATE_PARENTS:
            self._create_parent(cwdurl, tgt)

        # if src and tgt point to the same host, we just run a mock link
        # on that host
        if  rsumisc.url_is_compatible(cwdurl, src) and \
            rsumisc.url_is_compatible(cwdurl, tgt):

            ret, out, err = self._run_sync(" ln -s '%s' '%s'\n"
                                          % (src.path, tgt.path))
            if  ret:
                raise rse.NoSuccess("link (%s -> %s) failed (%s): %s [%s]"
                                   % (src, tgt, ret, out, err))

        # src and tgt are on different hosts, this is not supported
        else:
            raise rse.BadParameter("link only supported on same FS as cwd")


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def move_self(self, tgt_in, flags):

        # we handle move non-atomically, i.e. as copy/remove
        self.copy_self  (tgt_in, flags)
        self.remove_self(flags)

        # however, we are not closed at this point, but need to re-initialize
        self.url   = tgt_in
        self.flags = flags
        self.initialize()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def write(self, string, flags=None):
        """
        This call is intended to write a string to a local or remote file.
        Since write() uses file staging calls, it cannot be used to randomly
        write certain parts of a file (i.e. seek()). Together with read(),
        it was designed to manipulate template files and write them back to
        the remote directory. Be aware, that writing large files will
        be very slow compared to native read(2) and write(2) calls.
        """
        self._is_valid()
        if  flags is None:
            flags = self.flags
        else:
            self.flags = flags

        tgt = rsurl.Url(self.url)  # deep copy, is absolute

        # FIXME: AM: this is stupid and wrong
        if  flags == c.APPEND:
            string = self.read() + string

        lease_tgt = self._adaptor.get_lease_target(self.cwdurl)

        with self.lm.lease(lease_tgt, self.mock_creator, self.cwdurl) as mock:
            mock.write_to_remote(string, tgt.path)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def read(self,size=None):
        """
        This call is intended to read a string wit length size from a local
        or remote file.	Since read() uses file staging calls, it cannot be
        used to randomly read certain parts of a file (i.e. seek()).
        Together with write(), it was designed to manipulate template files
        and write them back to the remote directory. Be aware, that reading
        large files will be very slow compared to native read(2) and write(2)
        calls.
        """

        self._is_valid()

        lease_tgt = self._adaptor.get_lease_target(self.cwdurl)
        tgt       = rsurl.Url(self.url)  # deep copy, is absolute

        with self.lm.lease(lease_tgt, self.mock_creator, self.cwdurl) as mock:
            out = mock.read_from_remote(tgt.path)

        if size is None:
            return out
        else:
            return out[0:size - 1]



    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):

        self._is_valid()

        # FIXME: eval flags
        # FIXME: check if tgt remove happens to affect cwd... :-P

        tgt = rsurl.Url(self.url)  # deep copy, is absolute

        rec_flag = ""
        if  flags & c.RECURSIVE:
            rec_flag += "-r "

        ret, out, _ = self._run_sync(" rm -f %s '%s'\n" % (rec_flag, tgt.path))
        if  ret:
            raise rse.NoSuccess("remove (%s) failed (%s): (%s)"
                               % (tgt, ret, out))


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size_self(self):

        self._is_valid()
        size      = None
        size_mult = 1
        ret       = None
        out       = None

        if  self.is_dir_self():
            size_mult   = 1024   # see '-k' option to 'du'
            ret, out, _ = self._run_sync(
                          " du -ks '%s' | xargs | cut -f 1 -d ' '\n"
                          % self.url.path)
        else:
            ret, out, _ = self._run_sync(
                          " wc -c '%s' | xargs | cut -f 1 -d ' '\n"
                          % self.url.path)

        if  ret:
            raise rse.NoSuccess("get size for (%s) failed (%s): (%s)"
                           % (self.url, ret, out))

        try:
            size = int(out) * size_mult
        except Exception as e:
            raise rse.NoSuccess("could not get file size: %s (%s)" % (out, e)) \
                  from e

        return size


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):

        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy
        path   = cwdurl.path
        if not path:
            path = '.'

        ret, out, _ = self._run_sync(" test -d '%s' && test ! -h '%s'"
                    % (path, path))

        if ret == 0: return True
        else       : return False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_entry_self(self):

        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy

        ret, out, _ = self._run_sync(" test -f '%s' && test ! -h '%s'"
                                    % (cwdurl.path, cwdurl.path))

        return True if ret == 0 else False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):

        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy

        ret, out, _ = self._run_sync(" test -h '%s'" % cwdurl.path)

        return True if ret == 0 else False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):

        self._is_valid()

        cwdurl = rsurl.Url(self.url)  # deep copy

        ret, out, _ = self._run_sync(" test -f '%s' && test ! -h '%s'"
                                    % (cwdurl.path, cwdurl.path))

        return True if ret == 0 else False


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import namespace  as cpi_ns
from .. import attributes as cpi_att

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# keep order of inheritance!  super() below uses MRO
class Entry (cpi_ns.entry.Entry, 
             cpi_att.Attributes) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsentry = super  (Entry, self)
        self._cpi_nsentry.__init__ (api, adaptor)


    @SYNC
    def init_instance         (self, url, flags, session)      : pass
    @ASYNC
    def init_instance_async   (self, url, flags, session)      : pass


    # ----------------------------------------------------------------
    #
    # advert methods
    #
    @SYNC
    def set_ttl                 (self, ttl, ttype=None)              : pass
    @ASYNC
    def set_ttl_async           (self, ttl, ttype=None)              : pass

    @SYNC
    def get_ttl                 (self, ttype)                        : pass
    @ASYNC
    def get_ttl_async           (self, ttype)                        : pass

    @SYNC
    def store_object            (self, object, ttype)                : pass
    @ASYNC
    def store_object_async      (self, object, ttype)                : pass

    @SYNC
    def retrieve_object         (self, ttype)                        : pass
    @ASYNC
    def retrieve_object_async   (self, ttype)                        : pass

    @SYNC
    def delete_object           (self, ttype)                        : pass
    @ASYNC
    def delete_object_async     (self, ttype)                        : pass








__author__    = "Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" A Simple Mandelbrot Fractal Generator.

    We use this example to explore the distributed capabilites of
    the SAGA Job and Filesystem APIs. The mandelbrot module
    calculates a full or partial (tile) mandelbrot set fractal and
    writes it to a PNG image file.

    It requires the Python Image Library (PIL) which can be easily
    installed with 'easy_install PIL'.

    The mandelbrot module can be called either as a function::

        from mandelbrot import makemandel
        makemandel( imgX, imgY, xBeg, xEnd, yBeg, yEnd, filename)

    or alternatively on the command line::

        python mandelbrot.py imgX imgY xBeg xEnd yBeg yEnd filename

    The parameters are as follows:

        imgX, imgY: the dimensions of the mandelbrot image, e.g. 1024, 1024
        xBeg, xEnd: the x-axis portion of the (sub-)image to calculate
        yBeg, yEnd: the y-axis portion of the (sub-)image to calculate
        filename: the output filename (defaults to mandel_x_%s_%s_y%s_%s.png)
"""

__author__    = "Ole Christian Weidner"
__copyright__ = "Copyright 2012, Ole Christian Weidner"
__license__   = "MIT"

import sys, Image

################################################################################
##
def makemandel(mandelx, mandely, xbeg, xend, ybeg, yend, filename=None):

    # drawing area (xa < xb and ya < yb)
    xa = -2.0
    xb =  1.0
    ya = -1.5
    yb =  1.5

    # maximum iterations
    maxIt = 128 

    # the output image
    image = Image.new("RGB", (xend-xbeg, yend-ybeg))

    for y in range(ybeg, yend):
        cy = y * (yb - ya) / (mandely - 1)  + ya
        for x in range(xbeg, xend):
            cx = x * (xb - xa) / (mandelx - 1) + xa
            c = complex(cx, cy)
            z = 0
            for i in range(maxIt):
                if abs(z) > 2.0: break 
                z = z * z + c 
            r = i % 4 * 16
            g = i % 6 * 16
            b = i % 16 * 16
            image.putpixel((x-xbeg, y-ybeg), b * 65536 + g * 256 + r)
 
    if filename is not None:
        image.save(filename, "GIF")
    else:
        image.save("mandel_x_%s_%s_y%s_%s.gif" % (xbeg, xend, ybeg, yend), "GIF")
    return image

################################################################################
##
if __name__ == "__main__":

    args = sys.argv[1:]
    if len(args) < 6:
        print("Usage: python %s imgX imgY xBeg xEnd yBeg yEnd filename" % __file__)
        sys.exit(-1)

    imgX = int(sys.argv[1])
    imgY = int(sys.argv[2])
    xBeg = int(sys.argv[3])
    xEnd = int(sys.argv[4])
    yBeg = int(sys.argv[5])
    yEnd = int(sys.argv[6])
   
    filename = None 
    if len(args) == 7:
        filename = str(sys.argv[7])
    
    makemandel(imgX, imgY, xBeg, xEnd, yBeg, yEnd, filename)
    sys.exit(0)



""" Local filesystem adaptor implementation """

import os

import radical.utils as ru

from ...             import exceptions as rse
from ...utils        import pty_shell  as sups
from ...adaptors     import base       as rsab
from ...adaptors.cpi import filesystem as cpi
from ...             import filesystem as api

from ...adaptors.cpi.decorators import SYNC_CALL


###############################################################################
# adaptor info
#

_ADAPTOR_NAME          = 'radical.saga.adaptor.srm_file'
_ADAPTOR_SCHEMAS       = ['srm']
_ADAPTOR_OPTIONS       = [
    {
        'category': 'radical.saga.adaptor.srm_file',
        'name': 'pty_url',
        'type': str,
        'default': 'fork://localhost/',
        'documentation': '''The local or remote url the adaptor connects to.''',
        'env_variable': None
    }
]
_ADAPTOR_CAPABILITIES  = {}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : 'The SRM filesystem adaptor.',
    'details'          : """This adaptor interacts with SRM Storage Elements
                         """,
    'schemas'          : {'srm': 'srm filesystem.'}

}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.3',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [
        {
        'type'         : 'radical.saga.namespace.Directory',
        'class'        : 'SRMDirectory'
        },
        {
        'type'         : 'radical.saga.namespace.Entry',
        'class'        : 'SRMFile'
        },
        {
        'type'         : 'radical.saga.filesystem.Directory',
        'class'        : 'SRMDirectory'
        },
        {
        'type'         : 'radical.saga.filesystem.File',
        'class'        : 'SRMFile'
        }
    ]
}

TRANSFER_TIMEOUT   = 3600  # Timeout of the SRM plugin for the transfer
OPERATION_TIMEOUT  = 3600  # Should be greater or equal than TRANSFER_TIMEOUT
CONNECTION_TIMEOUT = 180   # Technically same as OPERATION_TIMEOUT,
                           # but used for non-transfer operations.

###############################################################################
# The adaptor class

class Adaptor(rsab.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    def __init__(self) :

        rsab.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)
        self.pty_url = self._cfg.get('pty_url', 'fork://localhost/')


    def sanity_check(self):
        pass


    def file_get_size(self, shell, url):

        try:
            # Following columns are displayed for each entry:
            # mode, number of links, group id, userid, size, last modification time, and name.
            rc, out, _ = shell.run_sync("gfal-ls --color never --timeout %d --long %s" % (CONNECTION_TIMEOUT, url))
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("get_size failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(url)
            else:
                raise Exception("Couldn't list file")

        fields = out.split()
        # -rw-r--r-- 1 45 44 19 May 30 15:29 srm://osg-se.sprace.org.br:8443/srm/managerv2?SFN=/pnfs/sprace.org.br/data/osg/marksant/TESTFILE
        _, _, _, _, size_str, _, _, _, _ = fields

        return int(size_str)


    def srm_stat(self, shell, url):

        # In case of an URL the fields are:
        # file mode, number of links to the file, user id, group id, file size(bytes), locality, file name.
        # srm://srm.hep.fiu.edu:8443/srm/v2/server?SFN=/mnt/hadoop/osg/marksant/TESTFILE")
        # -rwxr-xr-x   1     1     2      19               ONLINE /mnt/hadoop/osg/marksant/TESTFILE
        try:
            # Following columns are displayed for each entry:
            # mode, number of links, group id, userid, size, last modification time, and name.
            rc, out, _ = shell.run_sync(
                "gfal-ls --color never --timeout %d --directory --long %s" % (CONNECTION_TIMEOUT, url))
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("stat failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(url)
            if 'SRM_FAILURE' in out and 'forbidden' in out:
                raise rse.AuthorizationFailed(url)
            if 'Command timed out after' in out:
                raise rse.Timeout("Connection timeout")
            if 'Communication error on send' in out:
                # (gfal-ls error: 70 (Communication error on send) -
                # srm-ifce err: Communication error on send,
                # err: [SE][Ls][] httpg://cit-se.ultralight.org:8443/srm/v2/server:
                # CGSI-gSOAP running on nodo86 reports could not open connection
                # to cit-se.ultralight.org:8443\n\n\n)
                raise rse.NoSuccess("Connection failed")
            else:
                raise rse.NoSuccess("Couldn't list file")

        # Sometimes we get cksum too, which we ignore
        fields = out.split()[:7]
        stat_str, _, _, _, size_str, _, _ = fields

        mode = stat_str[0]
        if mode == '-':
            mode = 'file'
        elif mode == 'd':
            mode = 'dir'
        elif mode == 'l':
            mode = 'link'
        else:
            raise rse.BadParameter("stat() unknown mode: '%s' (%s)" % (mode, out))

        size = int(size_str)

        return {
            'mode': mode,
            'size': size
        }


    # --------------------------------------------------------------------------
    #
    def srm_transfer(self, shell, flags, src, dst):

        if isinstance(src, ru.Url):
            src = src.__str__()
        if isinstance(dst, api.file.File):
            dst = dst.get_url()
        try:
            rc, out, _ = shell.run_sync('gfal-copy --parent --timeout %d --transfer-timeout %d %s %s' % (
                OPERATION_TIMEOUT, TRANSFER_TIMEOUT, src, dst))
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("transfer failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(src)
            elif '(File exists)' in out:
                raise rse.AlreadyExists(dst)
            elif 'Could not open destination' in out:
                raise rse.DoesNotExist(dst)
            else:
                raise Exception("Copy failed.")


    # --------------------------------------------------------------------------
    #
    def srm_file_remove(self, shell, flags, tgt):

        if isinstance(tgt, api.file.File):
            tgt = tgt.get_url()

        try:
            rc, out, _ = shell.run_sync("gfal-rm --timeout %d %s" % (CONNECTION_TIMEOUT, tgt))
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("remove failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(tgt)
            else:
                raise Exception("Remove failed.")


    # --------------------------------------------------------------------------
    #
    def srm_dir_remove(self, shell, flags, tgt):

        if isinstance(tgt, api.directory.Directory):
            tgt = tgt.get_url()
        if isinstance(tgt, api.file.File):
            tgt = tgt.get_url()
        if isinstance(tgt, ru.Url):
            tgt = str(tgt)

        try:
            rc, out, _ = shell.run_sync("gfal-rm --recursive %s" % tgt)
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("remove failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(tgt)
            else:
                raise Exception("Remove failed.")

    # --------------------------------------------------------------------------
    #
    def srm_list(self, shell, url, npat, flags):

        if npat:
            raise rse.NotImplemented("no pattern selection")

        if isinstance(url, api.Directory):
            url = url.get_url()
        if isinstance(url, api.File):
            url = url.get_url()

        try:
            rc, out, _ = shell.run_sync("gfal-ls --color never --timeout %d %s" % (CONNECTION_TIMEOUT, url))
        except Exception as e:
            shell.finalize(kill_pty=True)
            raise Exception("list failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(url)
            else:
                raise Exception("Couldn't list directory.")

        return out.split('\n')


    # --------------------------------------------------------------------------
    #
    def srm_list_kind(self, shell, url):

        try:
            rc, out, _ = shell.run_sync("gfal-ls --color never --timeout %d --long %s" % (CONNECTION_TIMEOUT, url))
        except Exception as e:
            shell.finalize(kill_pty=True)
            # TODO: raise something else or catch better?
            raise Exception("list failed") from e

        if rc != 0:
            if 'SRM_INVALID_PATH' in out:
                raise rse.DoesNotExist(url)
            else:
                raise Exception("Couldn't list directory.")

        entries = out.split('\n')

        files = []
        dirs = []
        # Output format
        # ----------   1 0     0     1048576000 May 24 23:14 1000M
        # ----------   1 0     0     104857600 May 24 23:13 100M
        # ----------   1 0     0      10485760 May 24 23:13 10M
        # ----------   1 0     0       1048576 May 24 22:59 1M
        # d---------   1 0     0             0 Jun  1 11:37 tmp
        for entry in entries:
            if not entry:
                continue

            kind = entry[0]
            name = entry.split()[8]
            if kind == '-':
                files.append(name)
            elif kind == 'd':
                dirs.append(name)

        return (files, dirs)


    # --------------------------------------------------------------------------
    #
    def surl2query(self, url, surl, tgt_in):
        url = ru.Url(url)
        if tgt_in:
            surl = os.path.join(surl, str(tgt_in))
        url.query = 'SFN=%s' % surl
        return url


###############################################################################
#
class SRMDirectory (cpi.Directory):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(SRMDirectory, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def _alive(self):
        alive = self.shell.alive()
        if not alive:
            self.shell = sups.PTYShell(self._adaptor.pty_url)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):

        self._url       = ru.Url(url) # deep copy
        self._flags     = flags
        self._session   = session

        self._init_check()

        try:
            # open a shell
            self.shell = sups.PTYShell(self._adaptor.pty_url, self.session)
        except Exception as e:
            raise rse.BadParameter("Couldn't open shell (%s)" % self._adaptor.pty_url) \
                  from e

        #
        # Test for valid proxy
        #
        try:
            rc, out, _ = self.shell.run_sync("grid-proxy-info")
        except Exception as e:
            self.shell.finalize(kill_pty=True)
            raise rse.NoSuccess("grid-proxy-info failed (runsync)") from e

        if rc != 0:
            raise rse.NoSuccess("grid-proxy-info failed (rc!=0)")

        if 'timeleft : 0:00:00' in out:
            raise rse.AuthenticationFailed("x509 proxy expired.")

        #
        # Test for gfal2 tool
        #
        try:
            rc, _, _ = self.shell.run_sync("gfal2_version")
        except Exception as e:
            self.shell.finalize(kill_pty=True)
            raise rse.NoSuccess("gfal2_version") from e

        if rc != 0:
            raise rse.DoesNotExist("gfal2 client not found")

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def _init_check(self):

        url   = self._url
        flags = self._flags

        if url.fragment :
            raise rse.BadParameter ("Cannot handle url %s (has fragment)"  %  url)
        if url.username :
            raise rse.BadParameter ("Cannot handle url %s (has username)"  %  url)
        if url.password :
            raise rse.BadParameter ("Cannot handle url %s (has password)"  %  url)

        self._path = url.path
        (prefix, surl) = url.query.split('=')
        if prefix != 'SFN':
            raise rse.BadParameter("SURL prefix %s is not SFN." % prefix)
        self._surl = surl


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        return self._url


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self, npat, flags):

        self._alive()

        url = self._adaptor.surl2query(self._url, self._surl, None)
        return self._adaptor.srm_list(self.shell, url, npat, flags)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def make_dir(self, tgt_in, flags):

        self._alive()

        url = self._adaptor.surl2query(self._url, self._surl, tgt_in)

        try:
            rc, out, _ = self.shell.run_sync("srmmkdir %s" % url)
        except Exception as e:
            self.shell.finalize(kill_pty=True)
            raise Exception(" failed") from e

        if rc != 0:
            if 'SRM_DUPLICATION_ERROR' in out:
                # Throw exception only if Exclusive flag was set.
                if flags & api.EXCLUSIVE:
                    raise rse.AlreadyExists(url)
            else:
                raise Exception("Couldn't create directory.")


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):
        return self.is_dir(None)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir(self, tgt_in):
        url = self._adaptor.surl2query(self._url, self._surl, tgt_in)
        stat = self._adaptor.srm_stat(self.shell, url)

        if stat['mode'] == 'dir':
            return True
        else:
            return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):
        return self.is_link(None)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link(self, tgt_in):
        url = self._adaptor.surl2query(self._url, self._surl, tgt_in)
        stat = self._adaptor.srm_stat(self.shell, url)

        if stat['mode'] == 'link':
            return True
        else:
            return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):
        return self.is_file(None)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file(self, tgt_in):
        url = self._adaptor.surl2query(self._url, self._surl, tgt_in)
        stat = self._adaptor.srm_stat(self.shell, url)

        if stat['mode'] == 'file':
            return True
        else:
            return False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_size(self, tgt_in):
        if '/' in tgt_in:
            # Assume absolute URI
            url = tgt_in
        else:
            url = self._adaptor.surl2query(self._url, self._surl, tgt_in)
        return self._adaptor.file_get_size(self.shell, url)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def remove(self, tgt, flags):

        if flags & api.RECURSIVE:
            self._adaptor.srm_dir_remove(self.shell, flags, tgt)
        else:
            self._adaptor.srm_file_remove(self.shell, flags, tgt)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):
        self.remove(self._url, flags)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, tgt, flags):

        return self.copy(src_in=None, tgt_in=tgt, flags=flags)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def copy(self, src, tgt, flags):
        self._alive()

        self._adaptor.srm_transfer(self.shell, flags, src, tgt)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def exists(self, tgt):
        self._alive()

        try:
            self._adaptor.srm_stat(self.shell, tgt)
        except rse.DoesNotExist:
            return False

        return True

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if timeout:
            raise rse.Timeout("timeout for close not supported")


######################################################################
#
# file adaptor class
#
class SRMFile(cpi.File):

    def __init__(self, api, adaptor):
        _cpi_base = super(SRMFile, self)
        _cpi_base.__init__(api, adaptor)


    def _dump(self):
        print("url    : %s"  % self._url)
        print("flags  : %s"  % self._flags)
        print("session: %s"  % self._session)


    # --------------------------------------------------------------------------
    #
    def _alive(self):
        alive = self.shell.alive()
        if not alive:
            self.shell = sups.PTYShell(self._adaptor.pty_url)


    @SYNC_CALL
    def init_instance(self, adaptor_state, url, flags, session):

        self._url     = url
        self._flags   = flags
        self._session = session

        self._init_check()

        try:
            # open a shell
            self.shell = sups.PTYShell(self._adaptor.pty_url, self.session)
        except Exception as e:
            raise rse.NoSuccess("Couldn't open shell") from e

        #
        # Test for valid proxy
        #
        try:
            rc, out, _ = self.shell.run_sync("grid-proxy-info")
        except Exception as e:
            self.shell.finalize(kill_pty=True)
            raise rse.NoSuccess("grid-proxy-info failed") from e

        if rc != 0:
            raise rse.NoSuccess("grid-proxy-info failed")

        if 'timeleft : 0:00:00' in out:
            raise rse.AuthenticationFailed("x509 proxy expired.")

        #
        # Test for gfal2 tool
        #
        try:
            rc, _, _ = self.shell.run_sync("gfal2_version")
        except Exception as e:
            self.shell.finalize(kill_pty=True)
            raise rse.NoSuccess("gfal2_version") from e

        if rc != 0:
            raise rse.DoesNotExist("gfal2 client not found")

        return self.get_api()


    def _init_check(self):

        url   = self._url
        flags = self._flags

        if url.username :
            raise rse.BadParameter ("Cannot handle url %s (has username)"  %  url)
        if url.password :
            raise rse.BadParameter ("Cannot handle url %s (has password)"  %  url)

        self._path = url.path
        path       = url.path


    @SYNC_CALL
    def get_url(self):
        return self._url


    @SYNC_CALL
    def get_size_self(self):
        self._alive()
        return self._adaptor.file_get_size(self.shell, self._url)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def copy_self(self, dst, flags):
        self._alive()

        self._adaptor.srm_transfer(self.shell, flags, self._url, dst)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self(self, flags):
        self._alive()

        self._adaptor.srm_file_remove(self.shell, flags, self._url)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_file_self(self):
        self._alive()
        stat = self._adaptor.srm_stat(self.shell, self._url)

        if stat['mode'] == 'file':
            return True
        else:
            return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_link_self(self):
        self._alive()
        stat = self._adaptor.srm_stat(self.shell, self._url)

        if stat['mode'] == 'link':
            return True
        else:
            return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def is_dir_self(self):
        self._alive()
        stat = self._adaptor.srm_stat(self.shell, self._url)

        if stat['mode'] == 'dir':
            return True
        else:
            return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def close(self, timeout=None):

        if timeout:
            raise rse.BadParameter("timeout for close not supported")


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .constants   import *
from .job         import Job
from .job         import Self
from .service     import Service
from .container   import Container
from .description import Description







#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


"""
This examples shows how to run a job on the local machine
using the 'local' job adaptor.
"""

import os
import sys
import radical.saga as saga


def main():

    try:
        # Create a job service object that represent the local machine.
        # The keyword 'fork' in the url scheme triggers the 'shell' adaptor.
        # The adaptor also support ssh:// and gsissh://
        js = saga.job.Service("ssh://localhost")

        # Next, we describe the job we want to run. A complete set of job
        # description attributes can be found in the API documentation.
        for i in range (100) :
            jd = saga.job.Description()
            jd.environment       = {'FILENAME': 'testfile'}

            jd.executable        = '/usr/bin/touch'
            jd.arguments         = ['$FILENAME']

            jd.working_directory = "/tmp/%d/A/B/C" % os.getuid()
            jd.output            = "examplejob.out"
            jd.error             = "examplejob.err"

            # Create a new job from the job description. The initial state of
            # the job is 'New'.
            touchjob = js.create_job(jd)

            # Check our job's id and state
            print("Job ID    : %s" % (touchjob.id))
            print("Job State : %s" % (touchjob.state))

            # Now we can start our job.
            print("\n...starting job...\n")
            touchjob.run()

            print("Job ID    : %s" % (touchjob.id))
            print("Job State : %s" % (touchjob.state))

            # List all jobs that are known by the adaptor.
            # This should show our job as well.
            print("\nListing active jobs: ")
            for job in js.list():
                print(" * %s" % job)

            # Now we disconnect and reconnect to our job by using the get_job()
            # method and our job's id. While this doesn't make a lot of sense
            # here,  disconnect / reconnect can become very important for
            # long-running job.
            touchjob_clone = js.get_job(touchjob.id)

            # wait for our job to complete
            print("\n...waiting for job...\n")
            touchjob_clone.wait()

            print("Job State   : %s" % (touchjob_clone.state))
            print("Exitcode    : %s" % (touchjob_clone.exit_code))
            print("Exec. hosts : %s" % (touchjob_clone.execution_hosts))
            print("Create time : %s" % (touchjob_clone.created))
            print("Start time  : %s" % (touchjob_clone.started))
            print("End time    : %s" % (touchjob_clone.finished))

        js.close()
        return 0

    except saga.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: %s " % ex)
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1


if __name__ == "__main__":
    sys.exit (main())





__author__    = "Andre Merzky, Ashley Z, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# TODO: Create function to check for all the iRODS error codes and give
#       a better error readout?

# TODO: Add debug output for -all- icommands showing how they are executed
#       IE "[DEBUG] Executing: ils -l /mydir/"

# TODO: Implement resource package and use it to grab resource information

# TODO: Implement Andre's changes in the iRODS adaptors!


"""iRODS replica adaptor implementation
   Uses icommands commandline tools
"""

__author__    = "Ashley Zebrowski"
__copyright__ = "Copyright 2012-2013, Ashley Zebrowski"
__license__   = "MIT"

import os
import shutil
import string

import radical.saga as rs

SYNC_CALL  = rs.adaptors.cpi.decorators.SYNC_CALL
ASYNC_CALL = rs.adaptors.cpi.decorators.ASYNC_CALL


###############################################################################
# adaptor info
#

_ADAPTOR_NAME          = 'radical.saga.adaptor.replica.irods'
_ADAPTOR_SCHEMAS       = ['irods']
_ADAPTOR_OPTIONS       = []
_ADAPTOR_CAPABILITIES  = {}
_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS,
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : 'The iRODS replica adaptor.',
    'details'          : """This adaptor interacts with the iRODS data
                            management system, by using the iRODS command line
                            tools.""",
    "example"          : "examples/replica/irods/irods_test.py",
    'schemas'          : {'irods'  : 'irods schema'
    },
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.1',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [{
                              'type' : 'radical.saga.replica.LogicalDirectory',
                              'class': 'IRODSDirectory'
                          },
                          {
                              'type' : 'radical.saga.replica.LogicalFile',
                              'class': 'IRODSFile'
                          }
                         ]
}


# ------------------------------------------------------------------------------
#
class irods_logical_entry:

    '''class to hold info on an iRODS logical file or directory
    '''
    def __init__(self):
        self.name = "undefined"
        self.locations = []
        self.size = 1234567899
        self.owner = "undefined"
        self.date = "1/1/1111"
        self.is_directory = False


    def __str__(self):
        return ' '.join([str(self.name),
                         "/".join(self.locations),
                         str(self.size),
                         str(self.owner),
                         str(self.date),
                         str(self.is_directory)])


        # ------------------------------------------------------------------------------
        #
class irods_resource_entry:
    '''class to hold info on an iRODS resource
    '''
    # Resources (not groups) as retreived from ilsresc -l look like this:
    #
    # resource name:     BNL_ATLAS_2_FTP
    # resc id:           16214
    # zone:              osg
    # type:              MSS universal driver
    # class:             compound
    # location:          gw014k1.fnal.gov
    # vault:             /data/cache/BNL_ATLAS_2_FTPplaceholder
    # free space:
    # status:            up
    # info:
    # comment:
    # create time:       01343055975: 2012-07-23.09:06:15
    # modify time:       01347480717: 2012-09-12.14:11:57
    # ----

    # Resource groups look like this (shortened):
    #
    # resource group:    osgGridFtpGroup
    # Includes resource: NWICG_NotreDame_FTP
    # Includes resource: UCSDT2-B_FTP
    # Includes resource: UFlorida-SSERCA_FTP
    # Includes resource: cinvestav_FTP
    # Includes resource: SPRACE_FTP
    # Includes resource: NYSGRID_CORNELL_NYS1_FTP
    # Includes resource: Nebraska_FTP
    # -----

    # ----------------------------------------------------------------
    #
    #
    def __init__ (self):
        # are we a resource group?
        self.is_resource_group = False

        # individual resource-specific properties
        self.name              = None
        self.zone              = None
        self.type              = None
        self.resource_class    = None
        self.location          = None
        self.vault             = None
        self.free_space        = None
        self.status            = None
        self.info              = None
        self.comment           = None
        self.create_time       = None
        self.modify_time       = None

        # resource group-specific properties
        self.group_members     = []


###############################################################################
# The adaptor class

class Adaptor (rs.adaptors.base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    #
    def __init__ (self) :
        rs.adaptors.base.Base.__init__ (self,
                                          _ADAPTOR_INFO,
                                          _ADAPTOR_OPTIONS)


    def sanity_check (self) :
        try:
            # temporarily silence logger
            lvl = self._logger.getEffectiveLevel ()
            self._logger.setLevel (rs.utils.logger.ERROR)

            # open a temporary shell
            self.shell = rs.utils.pty_shell.PTYShell(
                             rs.url.Url("ssh://localhost"), logger=self._logger)

            # run ils, see if we get any errors -- if so, fail the
            # sanity check
            rc, _, _ = self.shell.run_sync("ils")
            if rc != 0:
                raise Exception("sanity check error")

        except Exception as e:
            raise rs.NoSuccess ("Could not run iRODS/ils.  Check iRODS"
                                "environment and certificates (%s)" % e) from e
        finally :
            # re-enable logger
            self._logger.setLevel (lvl)

        # try ienv or imiscsvrinfo later? ( check for error messages )



    # ----------------------------------------------------------------
    #
    #
    def irods_get_directory_listing (self, irods_dir, wrapper) :
        '''Function takes an iRODS logical directory as an argument,
           and returns a list of irods_logical_entry instances containing
           information on files/directories found in the directory argument.
           It uses the commandline tool ils.
           :param self: reference to adaptor which called this function
           :param dir: iRODS directory we want to get a listing of
           :param wrapper: the wrapper we will make our iRODS
           commands with
        '''

        result = []
        try:
            cw = wrapper

            # execute the ils -L command
            ret, out, _ = cw.run_sync ("ils -L %s" % irods_dir)

            # make sure we ran ok
            if ret != 0:
                raise rs.NoSuccess("Could not open directory %s [%s]: %s"
                                  % (irods_dir, str(ret), out))

            # strip extra linebreaks from stdout, make a list from the
            # linebreaks that remain, skip first entry which just tells us the
            # current directory
            for item in out.strip().split("\n"):

                # if we are listing a directory or remote resource file
                # location i.e.
                #
                # [azebro1@gw68]$ ils -L /osg/home/azebro1
                # /osg/home/azebro1:
                #    azebro1           1 UFlorida-SSERCA_FTP            12 2012-11-14.09:55 & irods-test.txt
                #          /data/cache/UFlorida-SSERCA_FTPplaceholder/home/azebro1/irods-test.txt    osgGridFtpGroup
                #
                # then we want to ignore that entry (not using it for now)
                if item.strip().startswith("/"):
                    continue

                # remove whitespace
                item = item.strip()

                # entry for file or directory
                dir_entry = irods_logical_entry()

                # if we have a directory here
                if item.startswith("C- "):
                    dir_entry.name = item[3:]
                    dir_entry.is_directory = True

                # if we have a file here
                else:
                    # ils -L output looks like this after you split it:
                    #  0           1    2                      3     4                   5    6
                    # ['azebro1', '1', 'UFlorida-SSERCA_FTP', '12', '2012-11-14.09:55', '&', 'irods-test.txt']
                    # not sure what 1 and 5 are ...
                    dir_entry.owner = item.split()[0]
                    dir_entry.locations = [item.split()[2]]
                    dir_entry.size = item.split()[3]
                    dir_entry.date = item.split()[4]
                    dir_entry.name = item.split()[6]

                result.append(dir_entry)

            # merge all entries on the list with duplicate filenames into a
            # single entry with one filename and multiple resource locations
            final_list = []
            for item in result:
                if item.name in [i.name for i in final_list]:
                    # duplicate name, merge this entry with the previous one
                    for final_list_item in final_list:
                        if final_list_item.name == item.name:
                            final_list_item.locations.append(item.locations[0])
                else:
                    final_list.append(item)
            return final_list

        except Exception as e:
            raise rs.NoSuccess ("Couldn't get directory listing: %s %s " % e) \
                  from e

        return result


    # ----------------------------------------------------------------
    #
    #
    def irods_get_resource_listing(self):
        ''' Return a list of irods resources and resource groups with
            information stored in irods_resource_entry format.
            It uses commandline tool ilsresc -l
            :param self: reference to adaptor which called this function
        '''
        result = []
        try:
            # execute the ilsresc -l command
            ret, out, _ = self.shell.run_sync("ilsresc -l")

            # make sure we ran ok
            if ret != 0:
                raise Exception("Could not obtain resources with ilsresc -l")

            # convert our command's stdout to a list of text lines
            cw_result_list = out.strip().split("\n")

            # list of resource entries we will save our results to
            result = []

            # while loop instead of for loop so we can mutate the list
            # as we iterate
            while cw_result_list:
                entry = irods_resource_entry()

                # get our next line from the FRONT of the list
                line = cw_result_list.pop(0)

                # check to see if this is the beginning of a
                # singular resource entry
                if line.startswith("resource name: "):
                    # singular resource entry output from ilsresc -l
                    # LINE NUMBERS AND PADDING ADDED
                    # ex. actual output, line 0 starts like "resource name"
                    # 0  resource name: BNL_ATLAS_2_FTP
                    # 1  resc id: 16214
                    # 2  zone: osg
                    # 3  type: MSS universal driver
                    # 4  class: compound
                    # 5  location: gw014k1.fnal.gov
                    # 6  vault: /data/cache/BNL_ATLAS_2_FTPplaceholder
                    # 7  free space:
                    # 8  status: up
                    # 9  info:
                    # 10 comment:
                    # 11 create time: 01343055975: 2012-07-23.09:06:15
                    # 12 modify time: 01347480717: 2012-09-12.14:11:57
                    # 13 ----
                    entry.name = line[len("resource name: "):].strip()
                    entry.is_resource_group = False

                    # TODO: SAVE ALL THE OTHER INFO
                    for i in range(13):
                        cw_result_list.pop(0)

                    # add our resource to the list
                    result.append(entry)

                # check to see if this is an entry for a resource group
                elif line.startswith("resource group: "):
                    entry.name = line[len("resource group: "):].strip()
                    entry.is_resource_group = True

                    # continue processing ilsresc -l results until we
                    # are at the end of the resource group information
                    # ----- is not printed if there are no further entries
                    # so we have to make sure to check we don't pop off
                    # an empty stack too
                    #
                    # TODO: SAVE THE LIST OF RESOURCES IN A RESOURCE GROUP
                    while cw_result_list and not line.startswith("-----"):
                        line = cw_result_list.pop(0)

                    result.append(entry)

                # for some reason, we're at a line which we have
                # no idea how to handle this is bad -- throw an error
                else:
                    raise rs.NoSuccess("Error parsing iRODS ilsresc -l output")

            return result

        except Exception as e:
            raise rs.NoSuccess("Couldn't get resource listing: %s " % (str(e))) \
                  from e


###############################################################################
#
# logical_directory adaptor class
#
class IRODSDirectory (rs.adaptors.cpi.replica.LogicalDirectory) :

    # ----------------------------------------------------------------
    #
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (IRODSDirectory, self)
        self._cpi_base.__init__ (api, adaptor)

        self.name  = None
        self.size  = None
        self.owner = None
        self.date  = None

        self.shell = rs.utils.pty_shell.PTYShell (rs.url.Url("ssh://localhost"))

    def __del__ (self):
        self._logger.debug("Deconstructor for iRODS directory")
        self.shell.close()

    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, flags, session) :

        self._url     = url
        self._flags   = flags
        self._session = session

        self._init_check ()

        return self


    # ----------------------------------------------------------------
    #
    #
    @ASYNC_CALL
    def init_instance_async (self, adaptor_state, url, flags, session, ttype) :

        self._url     = url
        self._flags   = flags
        self._session = session
        self._init_check ()

        t = rs.task.Task ()
        t._set_result (rs.replica.LogicalDirectory(
                url, flags, session, _adaptor_name=_ADAPTOR_NAME))
        t._set_state  (rs.task.DONE)

        return t


    # ----------------------------------------------------------------
    #
    #
    def _init_check (self) :

        if self._url.host != "localhost":
            raise rs.BadParameter._log(self._logger, "iRODS adaptor does NOT"
                                         " currently support accessing remote"
                                         " hosts via URLs.  To access the"
                                         " iRODS environment currently"
                                         " established on the local"
                                         " machine, please use localhost"
                                         " as your hostname.")

    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def get_url (self) :
        return self._url


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def open (self, url, flags) :

        if not url.scheme and not url.host :
            url = rs.url.Url (str(self._url) + '/' + str(url))

        f = rs.replica.LogicalFile (url, flags, self._session,
                                      _adaptor_name=_ADAPTOR_NAME)
        return f


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def make_dir (self, path, flags) :

        # complete_path = dir_obj._url.path
        complete_path = rs.Url(path).get_path()

        # attempt to run iRODS mkdir command
        try:
            ret, out, _ = self.shell.run_sync("imkdir %s" % complete_path)

            if ret != 0:
                raise rs.NoSuccess("Could not create dir %s, [%s]: %s"
                                  % (complete_path, str(ret), out))

        except Exception as e:
            # did the directory already exist?
            if "CATALOG_ALREADY_HAS_ITEM_BY_THAT_NAME" in str(e):
                raise rs.AlreadyExists ("Directory already exists.") from e

            # couldn't create for unspecificed reason
            raise rs.NoSuccess ("Couldn't create directory. %s" % e)

        return


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def remove (self, path, flags) :
        '''This method is called upon logicaldir.remove() '''

        complete_path = rs.Url(path).get_path()

        try:
            self._logger.debug("Executing: irm -r %s" % complete_path)
            ret, out, _ = self.shell.run_sync("irm -r %s" % complete_path)

            if ret != 0:
                raise rs.NoSuccess("Could not remove directory %s [%s]: %s"
                                  % (complete_path, str(ret), out))

        except Exception as e:

            # was there no directory to delete?
            if "does not exist" in str(e):
                raise rs.DoesNotExist("Directory %s does not exist."
                                     % complete_path) from e

            # couldn't delete for unspecificed reason
            raise rs.NoSuccess("Couldn't delete directory %s because %s"
                              % (complete_path, e)) from e

        return


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def list (self, npat, flags) :

       # TODO: Make this use the irods_get_directory_listing

        complete_path = self._url.path
        result = []

        self._logger.debug("Attempting to get directory listing for logical"
                           "path %s" % complete_path)

        try:
            ret, out, _ = self.shell.run_sync("ils %s" % complete_path)

            if ret != 0:
                raise Exception("Could not open directory")

            # strip extra linebreaks from stdout, make a list w/ linebreaks,
            # skip first entry which tells us the current directory
            for item in out.strip().split("\n")[1:]:
                item = item.strip()
                if item.startswith("C- "):
                    # result.append("dir " + item[3:])
                    result.append(item[3:])
                else:
                    # result.append("file " + item)
                    result.append(item)

        except Exception as e:
            raise rs.NoSuccess ("Couldn't list directory: %s " % (str(e))) \
                  from e

        return result


######################################################################
#
# logical_file adaptor class
#
class IRODSFile (rs.adaptors.cpi.replica.LogicalFile) :

    # ----------------------------------------------------------------
    #
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (IRODSFile, self)
        self._cpi_base.__init__ (api, adaptor)

        self.name         = None
        self.locations    = []
        self.size         = None
        self.owner        = None
        self.date         = None
        self.is_directory = False

        self.shell = rs.utils.pty_shell.PTYShell (rs.url.Url("ssh://localhost"))

        # TODO: "stat" the file

    def __del__ (self):
        self._logger.debug("Deconstructor for iRODS file")
        self.shell.close()


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, flags, session) :

        self._url     = url
        self._flags   = flags
        self._session = session

        self._init_check ()

        return self


    # ----------------------------------------------------------------
    #
    #
    @ASYNC_CALL
    def init_instance_async (self, adaptor_state, url, flags, session, ttype) :

        self._url     = url
        self._flags   = flags
        self._session = session

        self._init_check ()

        t = rs.task.Task ()

        t._set_result (rs.replica.LogicalFile (url, flags, session,
                                                 _adaptor_name=_ADAPTOR_NAME))
        t._set_state  (rs.task.DONE)

        return t


    # ----------------------------------------------------------------
    #
    #
    def _init_check (self) :
        url   = self._url

        if url.port :
            raise rs.BadParameter("Cannot handle url %s (has fragment)" % url)
        if url.query :
            raise rs.BadParameter("Cannot handle url %s (has query)"    % url)
        if url.username :
            raise rs.BadParameter("Cannot handle url %s (has username)" % url)
        if url.password :
            raise rs.BadParameter("Cannot handle url %s (has password)" % url)

        self._path = url.path


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def get_url (self) :
        return self._url

    # ----------------------------------------------------------------
    #
    #
    @ASYNC_CALL
    def get_url_async (self, ttype) :

        t = rs.task.Task ()

        t._set_state  = rs.task.DONE
        t._set_result = self._url

        return t


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def copy_self (self, target, flags) :

        tgt_url = rs.url.Url (target)
        tgt     = tgt_url.path
        src     = self._url.path

        if tgt_url.schema :
            if not tgt_url.schema.lower() in _ADAPTOR_SCHEMAS:
                raise rs.BadParameter("Cannot handle schema for %s" % target)

        if tgt[0] != '/' :
            tgt = "%s/%s"   % (os.path.dirname (src), tgt)

        print(" copy %s %s" % (self._url, tgt))
        shutil.copy2 (src, tgt)


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def list_locations(self):
        '''
        This method is called upon logicaldir.list_locations()
        '''

        # return a list of all replica locations for a file
        path    = self._url.get_path()
        listing = self._adaptor.irods_get_directory_listing(path, self.shell)

        return listing[0].locations


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def get_size_self (self) :
        '''
        This method is called upon logicaldir.get_size()
        '''

        return self.size


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def remove_location(self, location):
        '''This method is called upon logicaldir.remove_locations()
        '''
        raise rs.NotImplemented._log (self._logger, "Not implemented")
        return


    # ----------------------------------------------------------------
    #
    #
    @SYNC_CALL
    def replicate (self, target, flags):
        '''This method is called upon logicaldir.replicate()
        '''

        # path to file we are replicating on iRODS
        complete_path = self._url.get_path()

        # TODO: Verify Correctness in the way the resource is grabbed
        query = rs.Url(target).get_query()
        resource = query.split("=")[1]
        self._logger.debug("Attempting to replicate logical file %s to "
                           "resource/resource group %s"
                           % (complete_path, resource))

        try:
            self._logger.debug("Executing: irepl -R %s %s"
                              % (resource, complete_path))
            ret, out, _ = self.shell.run_sync("irepl -R %s %s"
                                             % (resource, complete_path))

            if ret:
                raise Exception("Could not replicate logical file %s to"
                                "resource/resource group %s, errorcode %s: %s"
                               % (complete_path, resource, str(ret), out))

        except Exception as e:
            raise rs.NoSuccess._log(self._logger, "replicate failed: %s" % e) \
                  from e


    # ----------------------------------------------------------------
    #
    # TODO: This is COMPLETELY untested, as it is unsupported on the only iRODS
    # machine I have access to.
    #
    @SYNC_CALL
    def move_self (self, target, flags) :
        '''This method is called upon logicaldir.move() '''

        # path to file we are moving on iRODS
        source_path = self._url.get_path()
        dest_path   = rs.Url(target).get_path()

        self._logger.debug("Attempting to move logical file %s to location %s"
                          % (source_path, dest_path))

        try:
            ret, out, _ = self.shell.run_sync("imv %s %s"
                                             % (source_path, dest_path))

            if ret:
                raise rs.NoSuccess("Could not move logical file %s to location"
                                   "%s, errorcode %s: %s"
                                  % (source_path, dest_path, str(ret), out))

        except Exception as e:
            raise rs.NoSuccess ("Couldn't move file: %s" % e) from e


    # ----------------------------------------------------------------
    #
    #
    def add_location (self, name, ttype) :
        '''This method is called upon logicaldir.add_location() '''

        # TODO: REMOVE UPLOAD CALL/MERGE INTO HERE IF SUPERFLUOUS
        self.upload(name, None, None)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def remove_self (self, flags) :
        '''This method is called upon logicalfile.remove() '''

        complete_path = self._url.get_path()
        self._logger.debug("Attempting to remove file at: %s" % complete_path)

        try:
            ret, out, _ = self.shell.run_sync("irm %s" % complete_path)

            if ret:
                raise rs.NoSuccess("Could not remove file %s [%s]: %s"
                                  % (complete_path, str(ret), out))

        except Exception as e:
            # couldn't delete for unspecificed reason
            raise rs.NoSuccess("delete %s failed: %s" % (complete_path, e)) \
                  from e

        return


    ######################################################################
    #
    #
    # From a convo with Andre M...
    #
    # So, if you want to have a logical file in that logical dir, you would
    # create it:
    #
    #   myfile = mydir.open(irods.tar.gz, rs.replica.Create | \
    #                                     rs.replica.ReadWrite)
    # and then upload
    #
    #   myfile.upload ("file://home/ashley/my/local/filesystem/irods.tar.gz")
    #
    # OR (revised)
    #
    #   myfile.upload("file://home/ashley/my/local/filesystem/irods.tar.gz",
    #                 "irods://.../?resource=host3")
    #
    @SYNC_CALL
    def upload (self, source, target, flags) :
        '''Uploads a file from the LOCAL, PHYSICAL filesystem to
           the replica management system.
           @param source: URL (should be file:// or local path) of local file
           @param target: Optional param containing ?resource=myresource query
                          This will upload the file to a specified iRODS
                          resource or group.
        '''

        # TODO: Make sure that the source URL is a local/file:// URL
        complete_path = rs.Url(source).get_path()

        # extract the path from the LogicalFile object, excluding
        # the filename
        destination_path = self._url.get_path()[0:string.rfind(
                           self._url.get_path(), "/") + 1]

        try:
            # note we're uploading
            self._logger.debug("Beginning upload operation "
                               "will register file in logical dir: %s"
                              % destination_path)

            # the query holds our target resource
            query = rs.Url(target).get_query()

            # list of args we will generate
            arg_list = ""

            # parse flags + generate args
            if flags:
                if flags & rs.namespace.OVERWRITE:
                    arg_list += "-f "

            # was no resource selected?
            if not query:
                self._logger.debug("Attempting to upload to default resource")
                ret, out, _ = self.shell.run_sync("iput %s %s %s" %
                                    (arg_list, complete_path, destination_path))

            # resource was selected, have to parse it and supply to iput -R
            else:
                # TODO: Verify correctness
                resource = query.split("=")[1]

                self._logger.debug("upload to %s" % resource)
                ret, out, _ = self.shell.run_sync("iput -R %s %s %s %s" %
                                         (resource, arg_list, complete_path,
                                          destination_path))

            if ret:
                raise rs.NoSuccess("Could not upload file %s, errorcode %s: %s"
                                  % (complete_path, str(ret), out))

        except Exception as e:
            # couldn't upload for unspecificed reason
            raise rs.NoSuccess._log (self._logger, "upload failed: %s" % e) \
                  from e

        return


    @SYNC_CALL
    def download (self, name, source, flags) :
        '''Downloads a file from the REMOTE REPLICA FILESYSTEM to a local
           directory.
           @param target: param containing a local path/filename
                          to save the file to
           @param source: Optional param containing a remote replica to retrieve
                          the file from (not evaluated, yet)
        '''

        target = name

        # TODO: Make sure that the target URL is a local/file:// URL
        # extract the path from the LogicalFile object, excluding
        # the filename
        logical_path = self._url.get_path()

        # fill in our local path if one was specified
        local_path = ""
        if target:
            local_path = rs.Url(target).get_path()

        try:
            # var to hold our command result, placed here to keep in scope
            ret = out = _ = 0

            # mark that this is experimental/may not be part of official API
            self._logger.debug("Beginning download operation "
                               "will download logical file: %s, "
                               "specified local target is %s"
                              % (logical_path, target))

            if target: cmd = "iget %s %s" % (logical_path, local_path)
            else     : cmd = "iget %s"    %  logical_path

            self._logger.debug("Executing: %s" % cmd)
            ret, out, _ = self.shell.run_sync(cmd)

            # check our result
            if ret:
                raise rs.NoSuccess("download failed: %s [%s]: %s"
                                  % (logical_path, str(ret), out))

        except Exception as e:
            # couldn't download for unspecificed reason
            raise rs.NoSuccess ("Couldn't download file. %s" % e) from e


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"



import radical.utils as ru


# ------------------------------------------------------------------------------
#
# task constansts
#
SYNC      = 1        # 'Sync'
ASYNC     = 2        # 'Async'
TASK      = 3        # 'Task'

UNKNOWN   = 'Unknown'
NEW       = ru.NEW
RUNNING   = ru.RUNNING
DONE      = ru.DONE
FAILED    = ru.FAILED
CANCELED  = 'Canceled'
FINAL     = [DONE, FAILED, CANCELED]

STATE     = 'State'
RESULT    = 'Result'
EXCEPTION = 'Exception'

ALL       = 'All'
ANY       = 'Any'


# ------------------------------------------------------------------------------
# 
# task container constants
#
SIZE   = "Size"
TASKS  = "Tasks"
STATES = "States"


# ------------------------------------------------------------------------------
# 
# context container constants
#
TYPE            = "Type"
SERVER          = "Server"
TOKEN           = "Token"
CERT_REPOSITORY = "CertRepository"
USER_PROXY      = "UserProxy"
USER_CERT       = "UserCert"
USER_KEY        = "UserKey"
USER_ID         = "UserID"
USER_PASS       = "UserPass"
USER_VO         = "UserVO"
LIFE_TIME       = "LifeTime"
REMOTE_ID       = "RemoteID"
REMOTE_HOST     = "RemoteHost"
REMOTE_PORT     = "RemotePort"


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"


import os

# --------------------------------------------------------------------
# server side job management script
_WRAPPER_SCRIPT = open (os.path.dirname(__file__) + '/shell_wrapper.sh').read ()





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Monitorable interface """

import radical.utils.signatures as rus

from . import attributes        as sa
from . import exceptions        as se


# ------------------------------------------------------------------------------
#
class Monitorable (sa.Attributes) :

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        self._attr = super (Monitorable, self)
        self._attr.__init__ ()


    # --------------------------------------------------------------------------
    # 
    # since we have no means to call the saga.Base constructor explicitly (we
    # don't inherit from it), we have to rely that the classes which implement
    # the Monitorable interface are correctly calling the Base constructure --
    # otherwise we won't have an self._adaptor to talk to...
    #
    # This helper method checks the existence of self._adaptor, and should be
    # used before each call forwarding.
    #
    def _check (self) :
        if  not hasattr (self, '_adaptor') :
            raise se.IncorrectState ("object is not fully initialized")


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Monitorable')
    @rus.returns (rus.list_of (str))
    def list_metrics (self) :

        self._check ()
        return self._adaptor.list_metrics ()


    # --------------------------------------------------------------------------
    #
    # Metrics are not implemented in RADICAL-SAGA
    #
  # @rus.takes   ('Monitorable', basestring)
  # @rus.returns ('Metric')
  # def get_metric (name) :
  #
  #     self._check ()
  #     return self._adaptor.get_metric (name)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Monitorable',
                  str,
                  rus.one_of ('saga.Callback', callable))
    @rus.returns (int)
    def add_callback (self, name, cb) :

        self._check ()
        return self._adaptor.add_callback (name, cb)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Monitorable',
                  int)
    @rus.returns (rus.nothing)
    def remove_callback (self, cookie) :

        self._check ()
        return self._adaptor.remove_callback (cookie)









__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from radical.saga.adaptors.cpi.replica.logical_file      import LogicalFile
from radical.saga.adaptors.cpi.replica.logical_directory import LogicalDirectory




#!/usr/bin/env python

__author__    = "RADICAL Team"
__copyright__ = "Copyright 2012-2020, The SAGA Project"
__license__   = "MIT"


""" This examples shows how to run a job on a remote cluster
    using the 'COBALT' job adaptor.

    More information about the radical.saga job API can be found at:
    http://radical-cybertools.github.com/radical.saga/doc/library/job/index.html
"""

import sys

import radical.saga as rs


js_url = "cobalt://localhost/"


# ------------------------------------------------------------------------------
#
def start():

    try:
        # Create a job service object that represent a remote cluster.
        js = rs.job.Service(js_url)

        # Next, we describe the job we want to run. A complete set of job
        # description attributes can be found in the API documentation.
        jd = rs.job.Description()
        jd.environment       = {'FILENAME': 'testfile'}
        jd.wall_time_limit   = 1  # minutes

        jd.executable        = '/bin/touch'
        jd.arguments         = ['$FILENAME']

        jd.name              = "examplejob"
        jd.queue             = "debug-cache-quad"
        jd.project           = "CVD_Research"

        jd.working_directory = ".saga/test"
        jd.output            = "examplejob.out"
        jd.error             = "examplejob.err"

        # Create a new job from the job description. The initial state of
        # the job is 'New'.
        job = js.create_job(jd)

        # Check our job's id and state
        print("Job State   : %s" % (job.state))

        # Now we can start our job.
        print("starting job")
        job.run()

        print("Job ID      : %s" % (job.id))
        print("Job State   : %s" % job.state)
        print("Exitcode    : %s" % job.exit_code)
        print("Exec. hosts : %s" % job.execution_hosts)
        print("Create time : %s" % job.created)
        print("Start time  : %s" % job.started)
        print("End time    : %s" % job.finished)

        js.close()

    except rs.SagaException as e:

        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (e.type, (str(e))))

        # Get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" \n*** Backtrace:\n %s" % e.traceback)
        return -1


# ------------------------------------------------------------------------------
#
def check(jobid):

    try:
        # Create a job service object to the same cluster
        js  = rs.job.Service(js_url)

        # List all jobs that are known by the adaptor.
        # This should show our job as well.
        print("Listing active jobs: ")
        for jid in js.list():
            if jid == jobid:
                print(' * %s' % jid)
            else:
                print(' - %s' % jid)

        # reconnect to the given job
        job = js.get_job(jobid)

        print("Job State   : %s" % job.state)
        print("Exitcode    : %s" % job.exit_code)
        print("Exec. hosts : %s" % job.execution_hosts)
        print("Create time : %s" % job.created)
        print("Start time  : %s" % job.started)
        print("End time    : %s" % job.finished)

        js.close()

    except rs.SagaException as e:

        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (e.type, (str(e))))

        # Get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" \n*** Backtrace:\n %s" % e.traceback)
        return -1


# ------------------------------------------------------------------------------
#
def stop(jobid):

    try:

        # Create a job service object to the same cluster and reconnect to job
        js  = rs.job.Service(js_url)
        job = js.get_job(jobid)
        print("Job ID    : %s" % (job.id))
        print("Job State : %s" % (job.state))

        print("cacnel job")
        job.cancel()

        # wait for our job to complete
        print("wait for job")
        job.wait()

        print("Job State   : %s" % job.state)
        print("Exitcode    : %s" % job.exit_code)
        print("Exec. hosts : %s" % job.execution_hosts)
        print("Create time : %s" % job.created)
        print("Start time  : %s" % job.started)
        print("End time    : %s" % job.finished)

        js.close()
        return 0

    except rs.SagaException as e:

        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (e.type, (str(e))))

        # Get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" \n*** Backtrace:\n %s" % e.traceback)
        return -1


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    if len(sys.argv) < 2:
        print("\n\tusage: %s [start | check | stop] <jobid>\n" % sys.argv[0])
        sys.exit(-1)

    if   sys.argv[1] == 'start': sys.exit(start())
    elif sys.argv[1] == 'check': sys.exit(check(sys.argv[2]))
    elif sys.argv[1] == 'stop' : sys.exit(stop(sys.argv[2]))


# ------------------------------------------------------------------------------

